
from .values import ValuesGroup
from .maps import MapsGroup
from .editors import EditorsGroup
from .relations import RelationsGroup
from .actions import ActionsGroup
from .ui_config import UIConfigGroup
from .injection import InjectionGroup
from .other import OtherFeaturesGroup


from kabaret import flow


class OtherFeaturesGroup(flow.Object):

    override_home_oid = flow.Label(
'''
<h3>Override Home oid</h3>
You can override the object to show as the "Home".
This can be usefull when your session only has one project, 
if you want to programmatically decide/restrict which project to show from the current user, or 
if you just want to use a custom "landing" flow with actions to navigate to other projects.

<pre>session.cmds.Flow.set_home_oid('/MyLandingPage')</pre>

Note that the real "Home" page will not be accessible anymore using standart GUI.
Also note that your flows can use this command and dynamically change the oid to use 
(set it to None to revert to default behavior).
''' 
)

    custom_flow_object = flow.Label(
'''
<h3>Custom Home Object</h3>
The default Home page is used to create and access projects. It's not the sexiest
because it's really generic. If you want to provide a better experience, you can
define a custom home flow for that.<br>
<br>
(This is what <i>kabaret.flow_button_pages</i>  uses to give you a nicer Home page
with big project buttons instead of the default projects Map, get it at
https://pypi.org/project/kabaret.flow-button-pages )

<pre>
# Import stuff you will need:
from kabaret.app.ui.gui.standalone import KabaretStandaloneGUISession
from kabaret.app.actors.flow import Flow
from kabaret.app.actors.flow.generic_home_flow import AbstractHomeRoot

# Define the flow to use as Home:
class MyHomeFlow(flow.Object):
    [blah...]
    [blah...]
    [blah...]

# Define a new HomeRoot using your home flow:
class MyHomeRoot(AbstractHomeRoot):
    Home = flow.Child(MyHomeFlow)  # ! it MUST be named "Home"

# In your session, override the Flow actor with one using your root:
class MySession(KabaretStandaloneGUISession):

    def _create_actors(self):
        # dont call the base class here since you're
        # defining the Flow yourself.
        Flow(self, MyHomeRoot)
<pre>
'''
)




from kabaret import flow



class Item(flow.Object):

    status = flow.Param('NYS').watched()

    def child_value_changed(self, child_value):
        if child_value is self.status:
            self.touch()

class BackLog(flow.DynamicMap):

    @classmethod
    def mapped_type(cls):
        return Item

    def mapped_names(self, page_num=0, page_size=None):
        return [ 'Item{}'.format(i) for i in range(10) ]

    def columns(self):
        return ['Name', 'Status']

    def _fill_row_cells(self, row, item):
        row['Name'] = item.name()
        row['Status'] = item.status.get()

class Filtered(flow.DynamicMap):

    _group = flow.Parent()

    @classmethod
    def mapped_type(cls):
        return Item

    def source_map(self):
        return self._group.get_backlog()

    def mapped_names(self, page_num=0, page_size=None):
        src = self.source_map()
        return [
            i.name()
            for i in src.mapped_items(page_num, page_size)
            if int(i.name()[-1])%2
        ]

    def get_mapped(self, name):
        return self.source_map().get_mapped(name)

    def columns(self):
        return ['Name', 'Status']

    def _fill_row_cells(self, row, item):
        row['Name'] = item.name()
        row['Status'] = item.status.get()

class SubGroup(flow.Object):

    _group = flow.Parent()

    another_view = flow.Child(Filtered)

    def get_backlog(self):
        return self._group.get_backlog()

class MapAsViewGroup(flow.Object):

    doc = flow.Label(
        'A DynamicMap can also be used as a "View" on another Map, '
        'i.e to filter a map or gather item from different parents...<br>'
        'In this example, the "backlog" contains item and the "filtered" '
        'shows only the even numbered items from the backlog.<br>'
        '<ul>'
        '<li>'
            'Double clicking an item in "filtered" opens the page of the '
            '"backlog" item.'
        '</li>'
        '<li>'
            'Changing the status of an item in "backlog" update both '
            'map editors.'
        '</li>'
        '<li>'
            'Changing the status of an item in "backlog" also update a '
            'map editor under another group (see sub_group).'
        '</li>'
        '</ul>'
    )

    def get_backlog(self):
        return self.backlog

    backlog = flow.Child(BackLog)
    filtered = flow.Child(Filtered)
    sub_group = flow.Child(SubGroup)


from .interface import Editor_Interface

from qtpy import QtWidgets, QtCore, QtGui

from six import text_type as unicode

from ..event_filters import IgnoreMouseButton


class TextAreaEditor(QtWidgets.QWidget, Editor_Interface):
    '''
    This editor lets you enter multiline text value.

    Editor Type Names:
        textarea

    Options:
        html:           bool (True)  - Automatically show html as rich text
        buttons_side:    top/bottom/left/right - defaults to right
    '''

    @classmethod
    def can_edit(cls, editor_type_name):
        '''
        Must be implemented to return True if the given editor_type_name
        matches this editor.
        '''
        return editor_type_name in ('textarea',)

    def __init__(self, parent, options):
        QtWidgets.QWidget.__init__(self, parent)
        Editor_Interface.__init__(self, parent)

        self._last_loaded_txt = None

        # :'( we should not assume parent has this, but I've been looking for
        # another way to have the text edit not too big for more time that I should :'(
        # parent.setFixedHeight(100)
        self.default_height = 100

        self.setLayout(QtWidgets.QBoxLayout(QtWidgets.QBoxLayout.LeftToRight))
        self.layout().setContentsMargins(1, 2, 1, 2)
        self.layout().setSpacing(1)

        self._te = QtWidgets.QTextEdit(self)
        self._te.viewport().installEventFilter(IgnoreMouseButton(self._te))

        self._apply_button = QtWidgets.QToolButton(self)
        self._apply_button.setText('Apply')
        self._apply_button.clicked.connect(self._apply_edit)
        self._apply_button.setShortcut(QtGui.QKeySequence("Ctrl+Return"))
        self._apply_button.setEnabled(False)
        self._apply_button.setVisible(False)

        self._cancel_button = QtWidgets.QToolButton(self)
        self._cancel_button.setText('Cancel')
        self._cancel_button.clicked.connect(self._on_cancel)
        self._cancel_button.setShortcut(QtGui.QKeySequence("Escape"))
        self._cancel_button.setEnabled(False)
        self._cancel_button.setVisible(False)

        self.layout().addWidget(self._te)
        self.buttons_layout = QtWidgets.QBoxLayout(QtWidgets.QBoxLayout.TopToBottom)
        self.buttons_layout.setContentsMargins(0,0,0,0)
        self.buttons_layout.setSpacing(0)
        self.buttons_layout.addWidget(self._apply_button)
        self.buttons_layout.addStretch(100)
        self.buttons_layout.addWidget(self._cancel_button)
        self.layout().addLayout(self.buttons_layout)

        self.apply_options(options)

    def sizeHint(self):
        return QtCore.QSize(self.default_height,100)

    def set_editable(self, b):
        '''
        Must be implemented to prevent editing if b is False.
        Visual cue should also be given to the user.
        '''
        self._te.setReadOnly(not b)
        self._apply_button.setVisible(b)
        self._cancel_button.setVisible(b)
        if b:
            self._te.textChanged.connect(self._on_edited)

    def set_button_side(self, side):
        d1, d2 = dict(
            bottom=(QtWidgets.QBoxLayout.TopToBottom, QtWidgets.QBoxLayout.LeftToRight), 
            top=(QtWidgets.QBoxLayout.BottomToTop, QtWidgets.QBoxLayout.LeftToRight),
            left=(QtWidgets.QBoxLayout.RightToLeft, QtWidgets.QBoxLayout.TopToBottom),
            right=(QtWidgets.QBoxLayout.LeftToRight, QtWidgets.QBoxLayout.TopToBottom),
        ).get(str(side).lower(), (None, None))
        if d1 is None:
            return self.set_button_side('right')

        self.layout().setDirection(d1)
        self.buttons_layout.setDirection(d2)

    def apply_options(self, options):
        '''
        Must be implemented to configure the editor as
        described by the options dict.
        '''
        self._te.setAcceptRichText(options.get('html', False))
        self.set_button_side(options.get('buttons_side'))

    def _set_buttons_enabled(self, b):
        self._apply_button.setEnabled(b)
        self._apply_button.setVisible(b)

        self._cancel_button.setEnabled(b)
        self._cancel_button.setVisible(b)

    def _on_cancel(self):
        self.update()

    def _apply_edit(self):
        self.apply()

    def update(self):
        '''
        Must be implemnented to show the value returned by self.fetch_value()
        Your code should call self._on_updated() at the end.
        '''
        txt = unicode(self.fetch_value() or '')
        self._last_loaded_txt = txt
        if not self._te.acceptRichText():
            self._te.setPlainText(txt)
        else:
            self._te.setHtml(txt)
        self._on_updated()

    def get_edited_value(self):
        '''
        Must be implemented to return the value currently displayed.
        '''
        value = self._te.toPlainText()
        return value

    def _show_edited(self):
        '''
        Must be implemented to show that the displayed value
        needs to be applied.
        '''
        if self.get_edited_value() == self._last_loaded_txt:
            self._show_clean()
            return
        self._set_buttons_enabled(True)
        self._te.setProperty('edited', True)
        self._te.setProperty('applying', False)
        # self.style().unpolish(self._te)
        self.style().polish(self._te)

    def _show_applied(self):
        '''
        Must be implemented to show that the displayed value
        as been saved.
        In a clean scenario, applying edits will trigger an update()
        and this state should disapear.
        If you are using the Editor without this kind of round trip,
        you can call update here.
        '''
        self._set_buttons_enabled(False)
        self._te.setProperty('edited', False)
        self._te.setProperty('applying', True)
        # self.style().unpolish(self._te)
        self.style().polish(self._te)

    def _show_clean(self):
        '''
        Must be implemented to show that the displayed value is
        up to date.
        '''
        self._set_buttons_enabled(False)
        self._te.setProperty('edited', False)
        self._te.setProperty('applying', False)
        # self.style().unpolish(self._te)
        self.style().polish(self._te)

    def _show_error(self, error_message):
        '''
        Must be implemented to show that the given error occured.
        '''
        self._set_buttons_enabled(False)
        self._te.setProperty('edited', False)
        self._te.setProperty('applying', False)
        self._te.setProperty('error', True)
        self.style().unpolish(self._te)
        self.style().polish(self._te)
        self.setToolTip('!!!\nERROR: %s' % (error_message,))


import time
import os
import shutil

from qtpy import QtWidgets, QtCore, QtGui

from .interface import Editor_Interface
import kabaret.app.resources

from .default_thumbnail import DEFAULT_THUMBNAIL

class WindowMover(QtCore.QObject):
    def __init__(self, widget):
        super(WindowMover, self).__init__(widget)
        self.window = widget.window()
        self.down_pos = None
        self.start_pos = None
        
    def eventFilter(self, obj, event):
        if event.type() == event.MouseButtonPress:
            self.down_pos = event.globalPos()
            self.start_pos = self.window.pos()
            obj.grabMouse()
            return True
        elif self.down_pos is not None and event.type() == event.MouseMove:
            delta = event.globalPos() - self.down_pos
            self.window.move(self.start_pos + delta)
            return True
        elif event.type() == event.MouseButtonRelease:
            self.down_pos = None
            self.start_pos = None
            obj.releaseMouse()
            return True
        else:
            return super(WindowMover, self).eventFilter(obj, event)


class ScreenGrabDialog(QtWidgets.QDialog):
    def __init__(self, parent, hide_window=True):
        super(ScreenGrabDialog, self).__init__(parent)
        
        if hide_window:
            self.window_to_hide = parent.window()
        else:
            self.window_to_hide = None
        self.restore_geom = None

        self._decoration_height = None
        self.winframe_w = 4
        self.winframe_h = 8
        
        self.setWindowTitle('SnapShot')
        
        self.screen_pix = self.grab_screen()
        
        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setContentsMargins(0,0,0,0)
        self.layout().setSpacing(0)
        
        self.view = QtWidgets.QGraphicsView(self)
        self.view.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.view.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.view.setResizeAnchor(self.view.NoAnchor)
        self.view.setTransformationAnchor(self.view.NoAnchor)
        self.view.setAlignment(QtCore.Qt.AlignTop|QtCore.Qt.AlignLeft)
        self.view.setCursor(QtCore.Qt.SizeAllCursor)
        self.layout().addWidget(self.view)
        
        self.mover = WindowMover(self)
        self.view.installEventFilter(self.mover)

        b = QtWidgets.QPushButton('Capture', self)
        b.clicked.connect(self.accept)
        self.layout().addWidget(b)
        
        self.scene = QtWidgets.QGraphicsScene()
        self.pix_item = self.scene.addPixmap(self.screen_pix)
        
        self.view.setScene(self.scene)

        self.resize(100, 100)
        
    def grab_screen(self):
        if self.window_to_hide is not None:
            self.restore_geom = self.window_to_hide.geometry()
            self.window_to_hide.hide()
            time.sleep(0.2)
        desktop = QtWidgets.QApplication.desktop()
        screen_pix = QtGui.QPixmap.grabWindow(
            desktop.winId(), 0, 0, desktop.width(), desktop.height()
        )        
        return screen_pix
    
    def moveEvent(self, event):
        pos = event.pos()
        x = pos.x()
        y = pos.y()
        if self._decoration_height is None:
            style = self.style()
            self._decoration_height = style.pixelMetric(style.PM_TitleBarHeight)
            self.winframe_w = 3
        self.pix_item.setPos(-x -self.winframe_w, -y -self._decoration_height -self.winframe_h)
        
    def get_croped_pix(self):
        return QtGui.QPixmap.grabWidget(self.view)
    
    def exec_(self):
        ret = super(ScreenGrabDialog, self).exec_()

        if self.window_to_hide is not None:
            self.window_to_hide.show()
            if self.restore_geom is not None:
                self.window_to_hide.setGeometry(self.restore_geom)
        
        return ret


class ThumbnailWidget(QtWidgets.QToolButton):

    def __init__(self, parent):
        super(ThumbnailWidget, self).__init__(parent)
        self._force_height = None
        self.maximum_height = 126

        self.path = None
        self.pix = QtGui.QPixmap()
        self._load()

        self.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)
        self.setAutoRaise(False)
        self.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        self.setArrowType(QtCore.Qt.NoArrow)

        # self.setFrameShape(QtWidgets.QFrame.StyledPanel)
        # self.setFrameShadow(QtWidgets.QFrame.Plain)
        # self.setLineWidth(2)
    
        self._menu = QtWidgets.QMenu()
        self._menu.addAction('Take a snapshot', self._on_snapshot_menu)
        self._menu.addAction('Copy path to clipboard', self._on_copy_path_to_clipboard)
        self.setMenu(self._menu)

    def force_height(self, height, maximum_height=None):
        if maximum_height:
            self.maximum_height = maximum_height
        self._force_height = height
        self._load()

    def set_path(self, path):
        self.path = path
        self._load()
            
    def _load(self):
        if self.path:
            self.pix.load(self.path)
        else:
            self.pix = QtGui.QPixmap(DEFAULT_THUMBNAIL)
        pix = self._get_scaled_pix()
        self.setIcon(QtGui.QIcon(pix))
        self.setIconSize(pix.size())
        self.setMinimumSize(pix.size())
    
    def _move_to_backup(self):
        bak_base = self.path+'.bak_'+str(time.time())
        bakpath = bak_base
        for i in range(10):
            if not os.path.exists(bakpath):
                break
            bakpath = bak_base+'_'+str(i)
        os.rename(self.path, bakpath)
    
    def update_from_path(self, path):
        '''
        Updates the content of the thumbnail using
        the given file.
        
        A copy of the current thumbnail file is made 
        before overwriting it.
        '''

        if not self.path:
            raise Exception('Cannot update thumnail: path not set')
        
        if not os.path.exists(path):
            raise Exception('Cannot update thumbnail from unexisting source %r.'%(path,))
                
        _, my_extension = os.path.splitext(self.path)
        _, in_extension = os.path.splitext(path)
        if my_extension != in_extension:
            raise Exception('Cannot, update thumbnail from %r: wrong extension.'%(path,))
        
        dir, base = os.path.split(self.path)
        if os.path.exists(self.path):
            self._move_to_backup()
        else:
            if not os.path.exists(dir):
                os.makedirs(dir)
        
        shutil.copy(path, self.path)
        self._load()

    def _on_copy_path_to_clipboard(self):
        cp = QtWidgets.QApplication.clipboard()
        cp.setText(self.path)
        
    def _on_snapshot_menu(self):
        self._do_snapshot(hide_window=False)
        
    def _on_hide_and_snapshot_menu(self):        
        self._do_snapshot()
        
    def _do_snapshot(self, hide_window=True):
        dialog = ScreenGrabDialog(self, hide_window=hide_window)
        result = dialog.exec_()
        if result == dialog.Accepted:
            pix = dialog.get_croped_pix()
            if pix is not None:
                self.update_from_pix(pix)
        dialog.deleteLater()
    
    def _get_scaled_pix(self):
        if self._force_height is None:
            if self.pix.height() > self.maximum_height:
                return self.pix.scaledToHeight(self.maximum_height, QtCore.Qt.SmoothTransformation)
            return self.pix
        return self.pix.scaledToHeight(min(self.maximum_height, self._force_height), QtCore.Qt.SmoothTransformation)

    def update_from_pix(self, pix):
        if not self.path:
            raise Exception('Cannot update thumbnail: path not set')
        
        dir, base = os.path.split(self.path)
        
        if os.path.exists(self.path):
            self._move_to_backup()
        else:
            if not os.path.exists(dir):
                os.makedirs(dir)
        
        pix.save(self.path)
        self._load()


class ThumbnailEditor(QtWidgets.QWidget, Editor_Interface):

    @classmethod
    def can_edit(cls, editor_type_name):
        return editor_type_name in ('thumbnail',)

    def __init__(self, parent, options):
        QtWidgets.QWidget.__init__(self, parent)
        Editor_Interface.__init__(self, parent)
        self._build()
        self.apply_options(options)

    def _build(self):
        self.setLayout(QtWidgets.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        self._tn = ThumbnailWidget(self)
        self.layout().addWidget(self._tn, 1, QtCore.Qt.AlignLeft)

        self._summary_lb = QtWidgets.QLabel(self)
        self._summary_lb.hide()
        self.layout().addWidget(self._summary_lb)

    def set_editable(self, b):
        pass

    def apply_options(self, options):
        pass

    def update(self):
        value = str(self.fetch_value() or '')
        self._tn.set_path(value)
        self._on_updated()

    def get_edited_value(self):
        return self._tn.path

    def _show_edited(self):
        self._tn.setProperty('edited', True)
        self._tn.setProperty('applying', False)
        self._tn.style().polish(self)

    def _show_applied(self):
        self._tn.setProperty('applying', True)

    def _show_clean(self):
        self._tn.setProperty('edited', False)
        self._tn.setProperty('applying', False)
        self._tn.style().polish(self)

    def _show_error(self, error_message):
        self._tn.setProperty('error', True)
        self._tn.style().polish(self)
        self.setToolTip('!!!\nERROR: %s' % (error_message,))



from qtpy import QtWidgets, QtCore, QtGui
import kabaret.app.resources
import kabaret.app.ui.gui.styles


class FormItem(QtWidgets.QTreeWidgetItem):
    '''
    Base class for all Form items (FormFields and internal stuff)

    Sublcass must implement the build() method to build the item widgets.
    '''

    _LAST_EXPANDED_STATES = {}
    _DEFAULT_EXPANDED_STATE = False

    def build(self):
        raise NotImplementedError()

    def level(self):
        p = self.parent()
        if p is None:
            return 0
        return p.level()+1

    def _pref_id(self):
        """ Returns an id used to manage preferences """
        fid = ''
        p = self.parent()
        if p is not None:
            fid = p._pref_id() + '.'
        fid += self.text(0)
        return fid

    def auto_expand(self):
        try:
            expanded = self.__class__._LAST_EXPANDED_STATES[self._pref_id()]
        except KeyError:
            expanded = self.__class__._DEFAULT_EXPANDED_STATE
        # if expanded:
        #     self.ensure_children_built()
        self.setExpanded(expanded)

    #FIXME: this triplet setExpanded/expanded/collapsed should be reduced to setExpanded only.
    def setExpanded(self, b):
        super(FormItem, self).setExpanded(b)
        self.__class__._LAST_EXPANDED_STATES[self._pref_id()] = b

    def expanded(self):
        self.__class__._LAST_EXPANDED_STATES[self._pref_id()] = True

    def collapsed(self):
        self.__class__._LAST_EXPANDED_STATES[self._pref_id()] = False

    def clicked(self, col):
        pass

    def activated(self, col):
        pass

    def apply_edit(self):
        pass

    def apply_all_edits(self):
        for i in range(self.childCount()):
            child = self.child(i)
            child.apply_all_edits()
        self.apply_edit()


class _PreloadItem(FormItem):
    '''
    The _PreloadItem is used by FormField to show a [+] (has children indicator) before
    having actually build the children.
    '''

    def __init__(self, parent):
        super(_PreloadItem, self).__init__(parent, ['_PRELOAD_ITEM_'])

    def build(self):
        '''Nothing to build here...'''
        pass

class FormField(FormItem):
    '''
    Base Class for all Form fields.
    FormField is responsible of managing its children:
        delete_all_children()
        build_children()
        ensure_children_built()

    The build_children() method must be implemented by subclasses.


    '''

    def __init__(self, parent):
        super(FormField, self).__init__(parent)
        self._preload_child = None
        self._children_built = False
        # self.setFlags(QtCore.Qt.NoItemFlags)

        if kabaret.app.ui.gui.styles.StylesManager.get().get_default_style().get_property('alternate_child_color', False):
            if self.level():
                color = self.parent().background(0).color()
                if self.level() % 2 == 0:
                    color = QtGui.QColor.fromHsl(color.hue(), color.saturation(), max(0, color.lightness()-5))
                    self.setBackground(0, color)
                else:
                    color = QtGui.QColor.fromHsl(color.hue(), color.saturation(), min(255, color.lightness()+10))
                    self.setBackground(0, color)
            else:
                self.setBackground(0, self.treeWidget().palette().color(QtGui.QPalette.Window))

        self.build()
        self.auto_expand()

    def setExpanded(self, b):
        if b:
            self._clear_preload_child()
        super(FormField, self).setExpanded(b)

    def expanded(self):
        self.ensure_children_built()
        super(FormField, self).expanded()

    def update_height(self, w):
        size = QtCore.QSize(32, 32).expandedTo(w.sizeHint())
        self.setSizeHint(1, size)
        self.treeWidget().scheduleDelayedItemsLayout()

    def show_may_have_children(self):
        if self._children_built:
            return
        if self._preload_child is None:
            self._preload_child = _PreloadItem(self)

    def _clear_preload_child(self):
        if self._preload_child is not None:
            index = self.indexOfChild(self._preload_child)
            c = self.takeChild(index)
            del c
            self._preload_child = None

    def delete_all_children(self):
        while self.childCount():
            c = self.takeChild(0)
            del c
        self._children_built = False

    def build_children(self):
        raise NotImplementedError()

    def ensure_children_built(self):
        if not self._children_built:
            self._clear_preload_child()
            self.build_children()
            self._children_built = True

    def wrap(self, w, box_direction=QtWidgets.QBoxLayout.LeftToRight, additionnal_widget=None):
        """
        :return: a widget with a QBoxLayout containing the widget w and the item name
        """
        tree = self.treeWidget()
        wrapper = QtWidgets.QWidget(tree)
        layout = QtWidgets.QBoxLayout(box_direction, wrapper)
        layout.setContentsMargins(0, 4, 0, 6)
        layout.setSpacing(0)
        hlayout = QtWidgets.QHBoxLayout()
        hlayout.setContentsMargins(2, 0, 0, 0)
        hlayout.setSpacing(10)

        title = QtWidgets.QWidget(wrapper)
        title.setContentsMargins(0, 0, 0, 0)
        icon_lbl = QtWidgets.QLabel()
        icon_lbl.setPixmap(self._get_config_icon().pixmap(QtCore.QSize(32, 32)))
        icon_lbl.setAlignment(QtCore.Qt.AlignVCenter)
        hlayout.addWidget(icon_lbl)
        title_lbl = QtWidgets.QLabel(self._label)
        title_lbl.setAlignment(QtCore.Qt.AlignVCenter)
        title_lbl.setWordWrap(True)
        hlayout.addWidget(title_lbl)
        if additionnal_widget:
            hlayout.addWidget(additionnal_widget)
        hlayout.addStretch()
        title.setLayout(hlayout)
        title.setMinimumSize(32, 32)

        layout.addWidget(title)
        layout.addWidget(w)
        wrapper.setLayout(layout)
        return wrapper

    def setItemWidget(self, col, w, alignment=QtCore.Qt.AlignLeft):
        tree = self.treeWidget()
        tree.setItemWidget(self, col, w)
        if col == 0:
            tree.setFirstItemColumnSpanned(self, True)
        self.setTextAlignment(col, alignment)


class FormTree(QtWidgets.QTreeWidget):

    def __init__(self, parent):
        super(FormTree, self).__init__(parent)
        self.setProperty('form_tree', True)

        self.setHeaderLabels(('Name', 'Value'))
        # self.setRootIsDecorated(False)
        self.setExpandsOnDoubleClick(False)

        self.itemClicked.connect(self._on_clicked)
        self.itemActivated.connect(self._on_activated)
        self.itemExpanded.connect(self._on_expanded)
        self.itemCollapsed.connect(self._on_collapsed)

        self.setVerticalScrollMode(self.ScrollPerPixel)
        self.setAutoScroll(False)

        h = self.header()
        h.setSectionResizeMode(0, self.header().ResizeToContents)
        h.hide()

        self.setIndentation(30)

        stylesheet = \
            '''
            QTreeView[form_tree="true"] {
                border: 0;
                outline: none;
                background: palette(window);
            }
            QTreeView[form_tree="true"]::item:selected {
                border: none;
            }
            QTreeView[form_tree="true"]::item:focus {
                border: none;
            }
            QTreeView[form_tree="true"]::item:hover {
                border: none;
                color: palette(highlight);
            }
            '''
        self.setStyleSheet(stylesheet)

        if kabaret.app.ui.gui.styles.StylesManager.get().get_default_style().get_property('alternate_child_color', False):
            self.drawRow = self.drawRow_alternate_child_color

    def _on_clicked(self, item, col):
        item.clicked(col)

    def _on_activated(self, item, col):
        item.activated(col)

    def _on_expanded(self, item):
        item.expanded()

    def _on_collapsed(self, item):
        item.collapsed()

    def keyPressEvent(self, event):
        '''
        Reimplemented to disable the enter key -> activate item
        '''
        if event.key() & QtCore.Qt.Key_Enter:
            event.accept()
        else:
            super(FormTree, self).keyPressEvent(event)

    def apply_all_edits(self):
        root = self.invisibleRootItem()
        for i in range(root.childCount()):
            item = root.child(i)
            item.apply_all_edits()

    def drawRow_alternate_child_color(self, *args, **kwargs):
        '''
        This is needed because Qt4 has a bug when overriding drawRow
        '''
        painter = args[0]
        option = args[1]
        index = args[2]
        item = self.itemFromIndex(index)

        margin_rect = QtCore.QRect(option.rect)
        margin_rect.setX((item.level() - 1.5) * self.indentation())
        margin_rect.setWidth(self.indentation())
        item_rect = QtCore.QRect(option.rect)
        item_rect.setX(margin_rect.right())

        parent = item.parent()
        if parent:
            painter.fillRect(item_rect, item.background(0))
            while parent:
                painter.fillRect(margin_rect, parent.background(0))
                margin_rect.moveLeft((parent.level() - 1.5) * self.indentation())
                parent = parent.parent()

        super(FormTree, self).drawRow(*args, **kwargs)



'''

    kabaret.flow.value_store

    Define the AbstractValueStore class (base for all value stores) and
    the MemoryValueStore class.

    A ValueStore is needed by Value instances to store and restore their state.

'''
import logging
import pprint
import six
from collections import OrderedDict


class AbstractValueStore(object):

    def __init__(self):
        super(AbstractValueStore, self).__init__()

    def log(self):
        logging.getLogger('kabaret.flow').info(pprint.pformat(self))

    def get(self, key):
        '''
        Must raise KeyError if not value exists for this key.
        '''
        raise NotImplementedError()

    def set(self, key, value):
        raise NotImplementedError()

    def delete(self, key):
        raise NotImplementedError()

    def incr(self, key, by=1):
        raise NotImplementedError(self)

    def decr(self, key, by=1):
        raise NotImplementedError()

    #--- Ordered String Set

    def oss_get(self, key):
        raise NotImplementedError()

    def oss_get_range(self, key, first, last):
        raise NotImplementedError()

    def oss_has(self, key, member):
        raise NotImplementedError()

    def oss_add(self, key, member, score):
        raise NotImplementedError()

    def oss_remove(self, key, member):
        raise NotImplementedError()

    def oss_len(self, key):
        raise NotImplementedError()

    def oss_get_score(self, key, member):
        raise NotImplementedError()

    def oss_set_score(self, key, member, score):
        raise NotImplementedError()

    #--- Hash

    def hash_get_key(self, key, hash_key):
        # hget
        raise NotImplementedError()

    def hash_has_key(self, key, hash_key):
        # hexists
        raise NotImplementedError()

    def del_hash_key(self, key, hash_key):
        # hdel
        raise NotImplementedError()

    def get_hash(self, key):
        # not the same as get_hash_as_dict bc it keep order
        raise NotImplementedError()

    def get_hash_as_dict(self, key):
        # hgetall
        raise NotImplementedError()

    def get_hash_keys(self, key):
        # hkeys
        raise NotImplementedError()

    def get_hash_len(self, key):
        # hlen
        raise NotImplementedError()

    def update_hash(self, key, mapping):
        # hmset
        raise NotImplementedError()

    def set_hash(self, key, mapping):
        raise NotImplementedError()

    def set_hash_key(self, key, hash_key, value):
        # hset
        raise NotImplementedError()


class MemoryValueStore(AbstractValueStore):

    def __init__(self):
        super(MemoryValueStore, self).__init__()
        self.store = {}

    def save_to(self, filename):
        with open(filename, 'w') as fh:
            fh.write('DATA = ' + pprint.pformat(self.store))
        logging.getLogger('kabaret.flow').info('MemoryValueStore saved to ' + filename)

    def load(self, filename):
        namespace = {}
        execfile(filename, namespace, namespace)
        data = namespace['DATA']
        self.store = data
        logging.getLogger('kabaret.flow').info('MemoryValueStore Loaded ' + filename)

    def log(self):
        logging.getLogger('kabaret.flow').info(pprint.pformat(self))

    def get(self, key):
        return self.store[key]

    def set(self, key, value):
        self.store[key] = value

    def delete(self, key):
        try:
            del self.store[key]
        except KeyError:
            return

    # deprecated, one must use hash
    # def update(self, key, **new_values):
    #     try:
    #         self.store[key]
    #     except KeyError:
    #         self.set(new_values)
    #     else:
    #         self.store[key].update(new_values)

    def incr(self, key, by):
        try:
            old_value = self.store[key]
        except KeyError:
            self.set(by)
        else:
            self.store[key] = old_value + by

    def decr(self, key, by):
        self.incr(-by)

    #--- Oredered String Set
    # stored as a very unefficient {string:score} 
    def oss_get(self, key):
        try:
            d = self.store[key]
        except KeyError:
            return []
        return [
            s
            for s, m in sorted(six.iteritems(d))
        ]

    def oss_get_range(self, key, first, last):
        return self.oss_get(key)[first:last]
        
    def oss_has(self, key, member):
        try:
            return member in self.store[key]
        except KeyError:
            return False

    def oss_add(self, key, member, score):
        self.oss_set_score(key, member, score)

    def oss_remove(self, key, member):
        try:
            d = self.store[key]
        except KeyError:
            d = {}
            self.store[key] = d
        del d[member]

    def oss_len(self, key):
        try:
            return len(self.store[key])
        except KeyError:
            return 0

    def oss_get_score(self, key, member):
        try:
            d = self.store[key]
        except KeyError:
            return None
        return d[member]

    def oss_set_score(self, key, member, score):
        try:
            d = self.store[key]
        except KeyError:
            d = {}
            self.store[key] = d
        d[member] = score

    #--- Hash
    # stored as a dict
    def hash_get_key(self, key, hash_key):
        # hget
        return self.store[key][hash_key]

    def hash_has_key(self, key, hash_key):
        # hexists
        try:
            return hash_key in self.store[key]
        except:
            return False

    def del_hash_key(self, key, hash_key):
        # hdel
        del self.store[key][hash_key]

    def get_hash(self, key):
        # not the same as get_hash_as_dict bc it keep order
        return (
            (k, self.hash_get_key(key, k))
            for k in self.get_hash_keys(key)
        )

    def get_hash_as_dict(self, key):
        # hgetall
        return dict(self.store[key])

    def get_hash_keys(self, key):
        # hkeys
        try:
            return self.store[key].keys()
        except:
            return []

    def get_hash_len(self, key):
        # hlen
        try:
            return len(self.store[key])
        except KeyError:
            return 0

    def update_hash(self, key, mapping):
        # hmset
        self.store[key].update(mapping)

    def set_hash(self, key, mapping):
        self.store[key] = OrderedDict(mapping)

    def set_hash_key(self, key, hash_key, value):
        # hset
        h = self.store.get(key, OrderedDict())
        h[hash_key] = value
        self.store[key] = h






import sys
import logging

def _tornado_import_object(name):
    '''code from tornado.util.import_object'''
    # py2 avoid unicodes:
    name = str(name)

    if name.count('.') == 0:
        return __import__(name, None, None)

    parts = name.split('.')
    obj = __import__('.'.join(parts[:-1]), None, None, [parts[-1]], 0)
    try:
        return getattr(obj, parts[-1])
    except AttributeError:
        raise ImportError("No module named %s" % parts[-1])


def import_object(
    object_qualified_name, reload_modules=True, clear_linecache=True
):
    # print('#------#')
    # print('#------# IMPORT OJECT', object_qualified_name)
    # print('#------#')
    if reload_modules:
        module_name = object_qualified_name.split('.', 1)[0]
        if 'KABARET' in module_name.upper():
            logging.getLogger('kabaret.flow').info('\n'.join((
                '[WARNING] CANNOT IMPORT PROJECT TYPE FROM KABARET',
                ' WITH reload_modules OPTION!'
            )))
        else:
            # print('#-----# reload modules', reload_modules, '->', module_name)
            for k in sorted(sys.modules.keys()):
                if k.startswith(module_name):
                    # print('[-]', k)
                    del sys.modules[k]
                else:
                    # print('[ ]', k)
                    pass

    if clear_linecache:
        # this is needed because of a bug in inspect which does not do
        # linecache.checkcache() in getsource()
        # fixed in py3k (I think)
        # It could be skipped for release... I guess. but needed for now
        # to have the flow run the reloaded code in client actions.
        import linecache
        linecache.clearcache()

    return _tornado_import_object(object_qualified_name)






import os

from qtpy import QtWidgets

from .. import Style


class NeonativeStyle(Style):

    def __init__(self):
        super(NeonativeStyle, self).__init__('Neonative')

    def apply(self, widget):
        app = QtWidgets.QApplication.instance()

        # --- No palette setup for neo-native style

        # --- Load and apply the css

        this_folder = os.path.dirname(__file__)
        css_file = os.path.join(this_folder, 'neonative_style.css')
        with open(css_file, 'r') as r:
            css = r.read()
        widget.setStyleSheet(css)





from __future__ import print_function

from .interface import Editor_Interface

from qtpy import QtWidgets, QtCore


class LabelEditor(QtWidgets.QLabel, Editor_Interface):
    '''
    This editor only show a static text.

    Editor Type Names:
        label

    Options:
        label:      the text to display, may be html
    '''

    @classmethod
    def can_edit(cls, editor_type_name):
        return editor_type_name == 'label'

    def __init__(self, parent, options):
        QtWidgets.QLabel.__init__(self, parent)
        Editor_Interface.__init__(self, parent)
        self.forced_text = None
        self.apply_options(options)
        self.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)

    def minimumSizeHint(self, *args, **kwargs):
        return super(LabelEditor, self).minimumSizeHint() * 1.1

    def sizeHint(self, *args, **kwargs):
        return super(LabelEditor, self).sizeHint() * 1.2

    def set_editable(self, b):
        pass

    def apply_options(self, options):
        self.setWordWrap(options.get("wrap", True))
        self.forced_text = options.get('text', None)

    def apply(self):
        pass

    def update(self, *args, **kwargs):
        if self.forced_text is not None:
            self.setText(self.forced_text)
        else:
            self.setText(self.fetch_value())
        self._on_updated()

    def _show_clean(self):
        pass

    def _show_error(self, error_message):
        print(error_message)
        self.setProperty('error', True)
        self.style().polish(self)
        self.setToolTip('!!! ERROR !!! {}'.format(error_message))


'''

    kabaret.flow.map

    Defines the Map Object which has a parametrable ordered list of children.

    The mapped children must derive from the type returned by mapped_type()
    Use the add() and remove() method in Action subclasses to modify the children list.


'''
import logging

from .exceptions import WIPException

from .exceptions import MissingChildError, MappedNameError
from .object import Object
from .relations import OrderedStringSetParam, HashParam
from .injection import injection

class DynamicMap(Object):
    '''
    A DynamicMap contains a procedural list of items.
    This list is computed every time its is needed.

    One must subclass this an implement the methods:
        mapped_names(self, page_num=0, page_size=None)
    and optionnaly:
        _get_mapped_item_type(self, mapped_name)

    '''

    @classmethod
    def mapped_type(cls):
        '''
        Returns the default (which is also be the base) type of the mapped objects.
        Subclasses may override this to specialize the children type.

        Default is to return Object.
        '''
        return Object

    def __init__(self, parent, name):
        super(DynamicMap, self).__init__(parent, name)

    def __len__(self):
        return len(self.mapped_names())

    def mapped_names(self, page_num=0, page_size=None):
        raise NotImplementedError()

    def has_mapped_name(self, name):
        return name in self.mapped_names()

    def mapped_items(self, page_num=0, page_size=None):
        names = self.mapped_names(page_num, page_size)
        ret = []
        for name in names:
            ret.append(self.get_mapped(name))
        return ret

    def columns(self):
        '''
        Returns the list of columns name.
        Subclasses may reimplement this and rows() to provide inline informations
        about the children.

        Default is to return ['Name'] which will contain the mapped object name in rows()
        '''
        return ['Name']

    def page_size(self):
        '''
        Subclasses must override this and current_page_num() to enable paging.
        '''
        return None

    def current_page_num(self):
        '''
        Subclasses must override this and page_size() to enable paging.
        '''
        return 0

    def rows(self):
        '''
        Returns a list of
        ::
            (oid, dict) 
        per child in the order given by mapped_names().
        All dicts must contains the keys found in self.columns() + an optional "_style" key.

        Subclasses should not reimplement this but _fill_row_cells() and _fill_row_style() to
        customize table cells for the mapped items.

        Default behavior is to return the 'name' keys only in rows.

        The _style value in the row must be a dict with those optional keys:
            - icon: the icon in first column
            - background-color: the whole row background-color
            - foreground-color: the row font color

            - <col_name>_<property>: set the <property> style in column <col_name> (example: 'name_icon')
        '''
        # TODO: implement another method for 'thumbnail' view
        # TODO: sort rows
        page_num = self.current_page_num()
        page_size = self.page_size()
        rows = []
        for item in self.mapped_items(page_num, page_size):
            rows.append(self.row(item))
        return rows

    def row(self, item):
        '''
        Returns one of the entry returned by rows:
        ::
            (item_oid, cells_data)
        '''
        row = {}
        self._fill_row_cells(row, item)
        style = {'icon': item.ICON}
        row['_style'] = style
        self._fill_row_style(style, item, row)

        return item.oid(), row

    def _fill_row_cells(self, row, item):
        '''
        Subclasses must override this to fill value for each column returned by columns()
        '''
        row['Name'] = item.name()

    def _fill_row_style(self, style, item, row):
        '''
        Subclasses must override this to configure the style of a given item.
        The 'row' dict contains values for all columns and may be used to decide
        of the style but not to be edited.

        The _style value in the row must be a dict with those optional keys:
            - icon: the icon in first column
            - background-color: the whole row background-color
            - foreground-color: the row font color

            - <col_name>_<property>: set the <property> style in column <col_name> (example: 'name_icon')
        '''
        pass

    def _get_mapped_item_type(self, mapped_name):
        '''
        Returns the type of the given mapped item.
        Default is to return the Map's default mapped type
        (with resolved type injection).

        NB: The GUI assumes that the returned type is a subclass
        of mapped_type().

        '''
        # MEMO:
        # To get a type from a 'fully qualified type name' string, with
        # a fallback to the map's default mapped type:
        # try:
        #     object_type = self._mng.qualified_type_name_to_type(
        #         object_qualified_type_name
        #     )
        # except ImportError:
        #     mapped_type = self.mapped_type()
        #     print(
        #         'WARNING: unable to access type %r. Downgrading to map default type %r' % (
        #             object_qualified_type_name, mapped_type
        #         )
        #     )
        #     object_type = mapped_type
        #
        # ALSO:
        # If you want to support mapped item type injection, you should use:
        #   mapped_type = injection.resolve(
        #       self.mapped_type(), self
        #       )
        # instead of just `self.mapped_type()`
        # Or call the default implementation which does exactly that
        return injection.resolve(
            self.mapped_type(), self
        )

    def _create_child(self, name, object_type):
        '''
        Creates and return a new child object of the given type.
        '''
        object = object_type(self, name)
        self._configure_child(object)
        return object

    def _configure_child(self, child):
        '''
        This can be used by subclass to modify mapped items right
        after their instanciation (by _create_child())

        Default is to do nothing.
        '''
        pass

    def __getitem__(self, name):
        return self.get_mapped(name)

    def get_mapped(self, name):
        '''
        Returns the item mapped under 'name', instanciating it if not yet done.
        '''
        try:
            return self._mng.get_existing_child(name)
        except MissingChildError:
            # This child does not exists yet
            # We will need to create and store it (if it is a registered mapped name)
            pass

        if not self.has_mapped_name(name):
            raise MappedNameError(self.oid(), name)

        object_type = self._get_mapped_item_type(name)
        return self._create_child(name, object_type)

# FIXME: UPDATE THIS IF STILL NEEDED:
    # def _bake_children_type(self):
    #     '''
    #     When the mapped_type's name has changed, every entry in self._items
    #     will have the wrong type stored.
    #     You can call this to update the whole list.

    #     Warning: this will instantiate all children!
    #     '''
    #     new_items = {}
    #     for name in self.mapped_names():
    #         child = self.get_mapped(name)
    #         object_qualified_type_name = child._mng.object_qualified_type_name()
    #         new_items[name] = object_qualified_type_name
    #     self._items.set(new_items)


class Map(DynamicMap):
    '''
    A Map manages a per-instance list of children Objects.

    '''
    _mapped_names = OrderedStringSetParam()
    _mapped_types = HashParam()

    def _item_cmp(self, name_a, name_b):
        '''
        This is used by bake_order.
        '''
        return cmp(name_a, name_b)

    def bake_order(self):
        '''
        Stores a new mapped names order according to the _item_cmp() method.
        This should not be useful but in edge cases like you loaded the order
        from an external source and did not get a satisfying order in mapped
        items.


        Note that this cannot express an order based on anything else than the
        mapped name.

        '''
        names = self.mapped_names()
        names.sort(cmp=self._item_cmp)  # /!\ is this py3 compatible ???
        mn = self._mapped_names
        for score, name in enumerate(names):
            mn.set_score(name, score)

    def mapped_names(self, page_num=0, page_size=None):
        if page_size is not None:
            first = (page_num * page_size)
            last = first + page_size
            return self._mapped_names.get_range(first, last)
        return self._mapped_names.get()

    def add(self, name, object_type=None):
        '''
        Adds an object to the map.
        If provided, object_type must be a subclass of the map's mapped_type (returned by the
        classmethod mapped_type())
        '''
        mapped_type = injection.resolve(
            self.mapped_type(), self
        )
        if object_type is None:
            object_type = mapped_type
        elif not issubclass(object_type, mapped_type):
            raise TypeError(
                'Cannot add %r of type %r to Map %r: not a subclass of %r' % (
                    name, object_type, self.oid(), mapped_type
                )
            )
        if '.' in name:
            raise TypeError(
                'Invalid object name %r (it must be a valid attribute name).' % (name,))
        try:
            exec(name + '=None') in {}
        except:
            raise TypeError(
                'Invalid object name %r (it must be a valid attribute name).' % (name,))
        if name in dir(self):
            raise ValueError(
                'Cannot add an item "%r", this name is already defined in the class "%s" (%s).' %
                (
                    name, self.__class__.__name__, self._mng.oid()
                )
            )
        if self._mapped_names.has(name):
            raise ValueError(
                'An item %r is already mapped in %r.' %
                (
                    name, self._mng.oid()
                )
            )

        object_qualified_type_name = self._mng.get_qualified_type_name(
            object_type)
        self._mapped_names.add(name, 0)
        self._mapped_types.set_key(name, object_qualified_type_name)

        return self.get_mapped(name)

    def remove(self, name):
        '''
        Removes an Object from the map.
        '''
        self._mapped_names.remove(name)
        self._mapped_types.del_key(name)

        try:
            self._mng.destroy_child(name)
        except MissingChildError:
            # was not yet instantiated, nothing to destroy
            pass

    def clear(self):
        '''
        Remove all Objects from the Map.
        '''
        item_names = self.mapped_names()

        self._mapped_names.revert_to_default()
        self._mapped_types.set({})

        for name in item_names:
            try:
                self._mng.destroy_child(name)
            except MissingChildError:
                # was not yet instantiated, nothing to destroy
                pass
        # The touch() is most of time not wanted yet since a
        # repopulation is probably going to occur.
        # So we let the user decide if he needs it:
        # self.touch()

    def _get_mapped_item_type(self, mapped_name):
        '''
        Returns the type of the given mapped item.
        NB: The GUI assumes that the returned type is a subclass
        of mapped_type().

        '''
        object_qualified_type_name = self._mapped_types.get_key(mapped_name)
        try:
            object_type = self._mng.qualified_type_name_to_type(
                object_qualified_type_name
            )
        except ImportError:
            mapped_type = injection.resolve(
                self.mapped_type(), self
            )
            logging.getLogger('kabaret.ui').debug(
                'WARNING: unable to access type %r. Downgrading to map default type %r' % (
                    object_qualified_type_name, mapped_type
                )
            )
            object_type = mapped_type
        return object_type


class RefsMap(DynamicMap):
    '''
    The RefsMaps lists all the references pointing to a given object
    (default is the RefMap's parent)

    '''

    def ref_of(self):
        '''
        Subclasses may override this to return the object this
        RefsMap lists references to.
        (default is to return this RefsMap's parent)
        '''
        return self._mng.parent

    def ref_source(self, ref):
        '''
        Subclasses may override this to return the object of interest 
        given the ref pointing to self.ref_of()
        (default is to return ref's parent)
        '''
        return ref._mng.parent

    def mapped_names(self, page_num=0, page_size=None):
        '''
        This will use the ref oids as mapped item name.
        Subclasses can override this and call base implementation
        to do a quick and cheap filtering based on ref oids.
        '''
        oids = self.ref_of()._mng.refs()
        return oids

    def get_mapped(self, name):
        '''
        '''
        if not self.has_mapped_name(name):
            raise MappedNameError(self.oid(), name)

        try:
            ref = self._mng.get_object(name)
        except ValueError:
            raise

        return self.ref_source(ref)


class OLD_HASH_Map(Object):

    # stores mapped name to (is_hidden, full_qualified_type_name), used for lazy instantiation
    _items = HashParam()

    @classmethod
    def mapped_type(cls):
        '''
        Returns the default (which is also be the base) type of the mapped objects.
        Subclasses may override this to specialize the children type.

        Default is to return Object.
        '''
        return Object

    def __init__(self, parent, name):
        super(Map, self).__init__(parent, name)

    def __len__(self):
        return self._items.len()

    def mapped_names(self):
        '''
        Ordered list of mapped names 
        '''
        return self._items.keys()

    def has_mapped_name(self, name):
        return self._items.has_key(name)

    # Should not be needed (use has_mapped_oid() instead)
    # def mapped_oids(self):
    #     '''
    #     Ordered list of mappend items oids.
    #     '''
    #     oid = self.oid()
    #     return [ oid+'/'+name for name in self.mapped_names() ]

    def has_mapped_oid(self, oid):
        my_oid = self.oid()
        if not oid.startswith(my_oid + '/'):
            return False
        name = oid[len(my_oid) + 1:]
        if '/' in name:
            # We could actually consider that a child of a mapped is inside this map, no ?
            return False
        return self._items.has_key(name)

    def mapped_items(self, page_num=0, page_size=None, include_hidden=False):
        names = self.mapped_names()
        if page_size:
            first_index = (page_num * page_size) + 1
            names = names[first_index:]

        found = 0
        ret = []
        for name in names:
            is_hidden, object_qualified_type_name = self._get_mapped_info(name)
            if is_hidden and not include_hidden:
                continue
            found += 1
            ret.append(self.get_mapped(name))
            if page_size and found > page_size:
                break
        return ret

    def columns(self):
        '''
        Returns the list of columns name.
        Subclasses may reimplement this and rows() to provide inline informations
        about the children.

        Default is to return ['Name'] which will contain the mapped object name in rows()
        '''
        return ['Name']

    def page_size(self):
        '''
        Subclasses must override this and current_page_num() to enable paging.
        '''
        return None

    def current_page_num(self):
        '''
        Subclasses must override this and page_size() to enable paging.
        '''
        return 0

    def show_hidden(self):
        '''
        Subclasses must override this to enable show_hidden toggle.
        '''
        return False

    def rows(self):
        '''
        Returns a list of (oid, dict) per child in the order given by mapped_names().
        All dicts must contains the keys found in self.columns() + an optional "_style" key.

        Subclasses should not reimplement this but _fill_row_cells() and _fill_row_style() to 
        customize table cells for the mapped items.

        Default behavior is to return the 'name' keys only in rows sorted by creation order.

        The _style value in the row must be a dict with those optional keys:
            - icon: the icon in first column
            - background-color: the whole row background-color
            - foreground-color: the row font color

            - <col_name>_<property>: set the <property> style in column <col_name> (example: 'name_icon')
        '''
        # TODO: implement another method for 'thumbnail' view
        page_num = self.current_page_num()
        page_size = self.page_size()
        show_hidden = self.show_hidden()
        rows = []
        for item in self.mapped_items(page_num, page_size, show_hidden):
            rows.append(self.row(item))
        return rows

    def row(self, item):
        '''
        Returns one of the entry returned by rows: (item_oid, cells_data)
        '''
        row = {}
        self._fill_row_cells(row, item)
        style = {'icon': item.ICON}
        row['_style'] = style
        self._fill_row_style(style, item, row)

        return item.oid(), row

    def _fill_row_cells(self, row, item):
        '''
        Subclasses must override this to fill value for each column returned by columns()
        '''
        row['Name'] = item.name()

    def _fill_row_style(self, style, item, row):
        '''
        Subclasses must override this to configure the style of a given item.
        The 'row' dict contains values for all columns and may be used to decide
        of the style but not to be edited.

        The _style value in the row must be a dict with those optional keys:
            - icon: the icon in first column
            - background-color: the whole row background-color
            - foreground-color: the row font color
            - activate_oid: the oid of a flow.Action to trigger on double-click

            - <col_name>_<property>: set the <property> style in column <col_name> (example: 'name_icon')
        '''
        pass

#---OLD
    def row_icons(self):
        '''
        Return a list of icon name per child.
        (in the order given gy mapped_names())
        '''
        raise WIPException('This is obsolete, use _style column in rows()')
        # icon_name_cache = {}
        # ret = []
        # items = self._items.get() # dont use self.mapped_names() to avoid a value store fetch for each child!
        # for name in sorted(items.keys()):
        #     qualified_type_name = items[name]
        #     try:
        #         icon_name = icon_name_cache[qualified_type_name]
        #     except KeyError:
        #         try:
        #             object_type = self._mng.qualified_type_name_to_type(qualified_type_name)
        #         except ImportError:
        #             icon_name = None
        #         else:
        #             icon_name = object_type.ICON
        #     icon_name_cache[qualified_type_name] = icon_name
        #     ret.append(icon_name)
        # return ret

    def add(self, name, object_type=None):
        '''
        Adds an object to the map.
        If provided, object_type must be a subclass of the map's mapped_type (returned by the
        classmethod mapped_type())
        '''
        # FIXME: This sucks, we need to update it on refactoring, plus subclasses cant protect their reserved
        # keywords :/
        # FIND SOMETHING BETTER !
        if name.startswith('_') or name in (
            'get_mapped', 'add', 'hide', 'remove', 'clear', 'mapped_type',
            'iteritems', 'mapped_names', 'columns', 'rows', 'row_icons',
        ):
            raise ValueError(
                'The name %r is reserved, cannot use it for an item name.' % (name,))

        mapped_type = self.mapped_type()
        if object_type is not None:
            if not issubclass(object_type, mapped_type):
                raise TypeError(
                    'Cannot add %r of type %r to Map %r: not a subclass of %r' % (
                        name, object_type, self.oid(), mapped_type
                    )
                )
        else:
            object_type = mapped_type

        try:
            self._mng.get_existing_child(name)
        except MissingChildError:
            # creating it will set it
            object = self._create_child(name, object_type)
            self._set_mapped_infos(object, hidden=False)
            return object
        else:
            raise ValueError('An item %r is already mapped in %r.' %
                             (name, self._mng.oid()))

    def set_hidden(self, name, hidden):
        mapped_item = self.get_mapped(name)
        self._set_mapped_infos(mapped_item, bool(hidden))
        mapped_item.touch()

    def remove(self, name):
        try:
            self._items.del_key(name)
        except ValueError:
            raise MissingChildError(self.oid(), name)

        try:
            self._mng.destroy_child(name)
        except MissingChildError:
            # was not yet instantiated, nothing to destroy
            pass

    def clear(self):
        item_names = self.mapped_names()
        self._items.set({})
        for name in item_names:
            try:
                self._mng.destroy_child(name)
            except MissingChildError:
                # was not yet instantiated, nothing to destroy
                pass
        # The touch() is most of time not wanted yet since a
        # repopulation is probably going to occur.
        # So we let the user decide if he needs it:
        # self.touch()

    # def _mapped_type_from_qualified_type_name(self, object_qualified_type_name):
    #     try:
    #         object_type = self._mng.qualified_type_name_to_type(object_qualified_type_name)
    #     except ImportError:
    #         mapped_type = self.mapped_type()
    #         print(
    #             'WARNING: unable to access type %r. Downgrading to map default type %r'%(
    #               object_qualified_type_name, mapped_type
    #             )
    #         )
    #         object_type = mapped_type
    #     return object_type

    def _get_mapped_info(self, mapped_name):
        info = self._items.get_key(mapped_name)
        is_hidden = info.startswith('!')
        object_qualified_type_name = info[1:]
        return is_hidden, object_qualified_type_name

    # FIXME: this mapped infos thing is crap.
    # -> we should use another related hash to store the is_hidden info
    # -> and this should be done in a subclass since not always needed.
    def _set_mapped_infos(self, mapped_item, hidden):
        object_qualified_type_name = mapped_item._mng.object_qualified_type_name()
        info = '%s%s' % (hidden and '!' or ' ', object_qualified_type_name)
        self._items.set_key(mapped_item.name(), info)

    def _create_mapped(self, name, object_qualified_type_name):
        try:
            object_type = self._mng.qualified_type_name_to_type(
                object_qualified_type_name
            )
        except ImportError:
            mapped_type = self.mapped_type()
            logging.getLogger('kabaret.flow').debug(
                'WARNING: unable to access type %r. Downgrading to map default type %r' % (
                    object_qualified_type_name, mapped_type
                )
            )
            object_type = mapped_type
        return self._create_child(name, object_type)

    def _create_child(self, name, object_type):
        '''
        Creates a child object of the given type, register it has attribute
        '''
        object = object_type(self, name)
        setattr(self, name, object)
        return object

    def get_mapped(self, name):
        '''
        Returns the item mapped under 'name', instanciating it if not yet done.
        '''
        if 1:
            try:
                return self._mng.get_existing_child(name)
            except MissingChildError:
                # This child does not exists yet
                # We will need to create and store it (if it is a registered mapped name)
                pass

#---OLD
        else:
            raise WIPException(
                ' Not sure why I was doing this instead of just looking in self._mng'
            )
            # #FIXME: this should not return classic attribute
            # try:
            #     # If something here exists by that name, let's hope it's a mapped item and return it
            #     return Object.__getattribute__(self, name)
            # except AttributeError:
            #     # This child does not exists (neither does an attribute of the same name :/ )
            #     # We will need to create and store it
            #     #print(err)
            #     pass

        if not self.has_mapped_name(name):
            # We use to do this:
            # raise MissingChildError(self.oid(), name)
            # but the child is not missing, it is actually not registered....
            # so:
            raise ValueError('No %r mapped item declared under %s' %
                             (name, self.oid()))
            # BN: Shouldn't this raise an Exception from flow.exceptions instead of ValueError ???

        is_hidden, object_qualified_type_name = self._get_mapped_info(name)
        return self._create_mapped(name, object_qualified_type_name)

#---OLD
    def xx__getattr__(self, name):
        raise WIPException(
            'I dont think I use that, nor I think it is a good idea')
        return self.get_mapped(name)

    def _bake_children_type(self):
        '''
        When the mapped_type's name has changed, every entry in self._items
        will have the wrong type stored.
        You can call this to update the whole list.

        Warning: this will instantiate all children!
        '''
        new_items = {}
        for name in self.mapped_names():
            child = self.get_mapped(name)
            object_qualified_type_name = child._mng.object_qualified_type_name()
            new_items[name] = object_qualified_type_name
        self._items.set(new_items)


'''


    - ComputedMap - 

    - Do we force MemoryValueStore ? or do we have two flavors ?
    - New names: ComputedMap and CachedComputedMap
        (the only diff should be the _items config, and the action to refresh)
    - Refactor to match the modifications in Map, and clean it up.



'''


# class RefreshDynamicMapAction(Action):

#     _map = Parent()

#     def get_buttons(self):
#         return self.oid(), [] # no buttons => no dialog, direct exec

#     def run(self, button):
#         self.message.set('Refreshing')
#         self._map._refresh(self.show_progress)
#         self.message.set('Done.')

#     def show_progress(self, nb_step, curr_step, msg):
#         self.message.set('Refreshing: %s (%s/%s'%(nb_step, curr_step, msg))

# class DynamicMap(Map):

#     refresh = Child(RefreshDynamicMapAction)

#     _items = Computed(dict) # overrides base class relation by creating a one with same name

#     @classmethod
#     def _create_value_store(cls, parent, name):
#         return MemoryValueStore()

#     def __init__(self, *args, **kwargs):
#         super(DynamicMap, self).__init__(*args, **kwargs)
#         self._refreshing = False
#         self._refresh(self.refresh.show_progress)

#     def child_touched(self, child):
#         '''
#         !!! Subclasses MUST call this base implementation if they override !!!
#         '''
#         super(DynamicMap, self).child_touched(child)

#     def compute_child_value(self, child_value):
#         if child_value == self._items:
#             if not self._refreshing:# and self._auto_refresh:
#                 self._refresh(self.refresh.show_progress)
#         else:
#             super(DynamicMap, self).compute_child_value(child_value)

#     def _refresh(self, progress=None):
#         self._refreshing = True

#         progress = progress or (lambda nb_step, curr_step, msg: None)

#         self.clear()

#         entries = self._fetch_entries(progress)

#         nb = len(entries)
#         for i, (name, data) in enumerate(entries.iteritems()):
#             progress(nb, i, 'Creating %r'%(name,))
#             #print('--->', (name,))
#             item = self.add(name, self._get_child_type(name, data))
#             self._configure_child(item, data)

#         self._refreshing = False
#         self._set_clean()
#         self.touch()

#     def _fetch_entries(self, progress):
#         '''
#         Subclasses must implement this to return a dict of child_name:child_data

#         The child_infos will later be passed to self._configure_child().
#         '''
#         raise NotImplementedError()

#     def _get_child_type(self, name, data):
#         '''
#         Subclass may override this to refine each new child type.
#         Default is to use mapped_type()
#         '''
#         return self.mapped_type()

#     def _configure_child(self, child, data):
#         '''
#         Subclasses must implement this to configure a newly created child.

#         The data arguments is the one returned for this child by _fetch_entries().
#         '''
#         raise NotImplementedError()


# class RefreshDirtyMapAction(Action):

#     _map = Parent()

#     def get_buttons(self):
#         return self.oid(), [] # no buttons => no dialog, direct exec

#     def run(self, button):
#         self.message.set('Forcing Refresh.')
#         self._map._set_clean()
#         self._map.touch()
#         self.message.set('Done.')

#     def show_progress(self, nb_step, curr_step, msg):
#         self.message.set('Refreshing: %s (%s/%s'%(nb_step, curr_step, msg))


# class DirtyMap(Map):
#     '''
#     A DirtyMap updates it content every time it is read (by calling the rows() method).
#     (In a similarly way a Dirty relation updates at every value get)

#     Subclasses MUST call self._refresh() at the beginning of rows() when overriding it.
#     '''

#     refresh = Child(RefreshDirtyMapAction)

#     @classmethod
#     def _create_value_store(cls, parent, name):
#         return MemoryValueStore()

#     def __init__(self, *args, **kwargs):
#         self._refreshing = False
#         super(DirtyMap, self).__init__(*args, **kwargs)
#         self._loaded = False

#     def rows(self):
#         '''
#         BEWARE: subclasses overridding this MUST call self._refresh(self.refresh.show_progress)
#         before accessing the items to fetch the rows data.
#         '''
#         self._refresh(self.refresh.show_progress)
#         return super(DirtyMap, self).rows()

#     def _refresh(self, progress=None):
#         self._refreshing = False#True
#         self._loaded = True

#         progress = progress or (lambda nb_step, curr_step, msg: None)

#         self.clear()

#         entries = self._fetch_entries(progress)

#         nb = len(entries)
#         for i, (name, data) in enumerate(entries.iteritems()):
#             progress(nb, i, 'Creating %r'%(name,))
#             #print('--->', (name,))
#             item = self.add(name, self._get_child_type(name, data))
#             self._configure_child(item, data)

#         self._refreshing = False

#     def _fetch_entries(self, progress):
#         '''
#         Subclasses must implement this to return a dict of child_name:child_data

#         The child_infos will later be passed to self._configure_child().
#         '''
#         raise NotImplementedError()

#     def _get_child_type(self, name, data):
#         '''
#         Subclass may override this to refine each new child type.
#         Default is to use mapped_type()
#         '''
#         return self.mapped_type()

#     def _configure_child(self, child, data):
#         '''
#         Subclasses must implement this to configure a newly created child.

#         The data arguments is the one returned for this child by _fetch_entries().
#         '''
#         raise NotImplementedError()

#     def get_mapped(self, name):
#         if not self._loaded:
#             print('!!!!!!!!!!!!!!')
#             print('!!!!!!!!!!!!!! AUTO FIRST REFRESH S', self._items.get())
#             print('!!!!!!!!!!!!!!')
#             self._refresh()
#             print('!!!!!!!!!!!!!! AUTO FIRST REFRESH E', self._items.get())
#         return super(DirtyMap, self).get_mapped(name)
#





import sys

from ...session import KabaretSession

from qtpy import QtCore, QtWidgets
from .widgets.main_window import MainWindowManager
# from .widgets.flow.flow_view import FlowView
from .widgets.flow import FlowViewPlugin

class KabaretStandaloneGUISession(KabaretSession):

    def __init__(self, session_name='Standalone', tick_every_ms=10, debug=False):
        super(KabaretStandaloneGUISession, self).__init__(session_name, debug)
        self.main_window_manager = None
        self.tick_every_ms = tick_every_ms

    def is_gui(self):
        return True

    def register_plugins(self, plugin_manager):
        super(KabaretStandaloneGUISession, self).register_plugins(plugin_manager)
        plugin_manager.register(FlowViewPlugin, 'kabaret.flow_view')

    def create_window_manager(self):
        return MainWindowManager.create_window(self)
        
    def start(self):
        app = QtWidgets.QApplication(sys.argv)
        QtWidgets.QApplication.setOrganizationName("Supamonks")
        QtWidgets.QApplication.setApplicationName("Kabaret")
        self.main_window_manager = self.create_window_manager()
        self.main_window_manager.install()

        timer = QtCore.QTimer(self.main_window_manager.main_window)
        timer.timeout.connect(self.tick)
        timer.start(self.tick_every_ms)

        app.exec_()

    def register_view_types(self):
        '''
        Subclasses can register view types and create defaults view here.
        Use:
            type_name = self.register_view_type(MyViewType)
        to register a view type.
        
        And optionally:
            self.add_view(type_name)
        to create a default view.
        '''
        super(KabaretStandaloneGUISession, self).register_view_types()

    def _get_layout_state(self):
        return self.main_window_manager.get_layout_state()

    def _set_layout_state(self, state):
        self.main_window_manager.set_layout_state(state)

    def _on_cluster_connected(self):
        '''
        Overridden to also dispatch a 'cluster_connection' event to GUI.
        '''
        super(KabaretStandaloneGUISession, self)._on_cluster_connected()
        self.dispatch_event(
            'cluster_connection',
            cluter_name=self._cluster_actor.get_cluster_name()
        )

    # def dispatch_event(self, event_type, **data):
    #     if self.main_window_manager is not None:
    #         self.main_window_manager.receive_event(event_type, data)
    #     super(KabaretStandaloneGUISession, self).dispatch_event(event_type, **data)




from kabaret.flow import (
    values, group,
    Object, Map, Action, ConnectAction,
    ChoiceValueSetAction, ChoiceValueSelectAction,
    Child, Parent, Computed, Connection,
    Param, IntParam, BoolParam, HashParam,
    Label, Separator,
)
from kabaret import flow

class ParentExampleChild(Object):

    my_parent = Parent()
    relations = Parent(2)


class ParentExample(Object):

    relations = Parent()
    my_child = Child(ParentExampleChild)

    doc = Label(
        '<font color=#F88>/!\\ BEWARE:</font>'
        ' When your Objects need to access other objects, it is better to use a '
        'Connection or a method returning the object to use.<br>'
        'By doing so, you make your Object more reusable.'
        'The Parent() relation is usefull to add a button to the GUI, or when the relation '
        'is strong enough to not need an overrid in subclasses...'

    )


class MyChild(Object):

    doc = Label(
        'Some pretty dumb Object...'
    )

    def summary(self):
        return 'This is a child object. Double click its name to enter it, double click here to open it.'


class Option(Object):

    param = Param('This param is inside the Connection target !')

    @classmethod
    def get_source_display(cls, oid):
        '''
        This is used by the GUI to display the source of a Connection.
        Default is to show the whole oid. Here we cut to the source label.
        '''
        return oid.split('/')[-1].replace('_', ' ').title()

#---- Relative Example

class Stuff(flow.Object):
    param = flow.Param("This is stuff 1")

class ActionStuff(flow.Action):
    def allow_context(self, context):
        return 'inline' not in context

class MapStuff(flow.Map):
    pass
    
class REStuff(flow.Object):

    sub_stuff = flow.Child(Stuff)
    an_action = flow.Child(ActionStuff)
    a_map = flow.Child(MapStuff)
    
class DynRelativeExample(flow.Relative):

    def __init__(self):
        super(DynRelativeExample, self).__init__(None)

    def get_relative_oid(self, owner):
        return owner._dyn_relative_oid.get()

class SelectRelative(flow.Action):
    
    _parent = flow.Parent()

    def needs_dialog(self):
        return True
    
    def get_buttons(self):
        self.message.set('Select the object to display and reload the page')
        return ['sub_stuff', 'an_action', 'a_map', 'an_error']
    
    def run(self, button):
        self._parent._dyn_relative_oid.set('some_stuff/'+button)

class RelativeExample(flow.Object):

    relative_doc = Label(
        'The group "some_stuff" contains an Object, an Action and a Map.<br>'
        'The next field shows the "sub_suff" inside "some_stuff" like if it was a Child of this Object.<br>'
        'Third field is the same with overidden label and icon.<br>'
        'The "Select" action let you select the object to display as the last relation (need a refresh).'
    )

    some_stuff = flow.Child(REStuff)
    relative = flow.Relative('some_stuff/sub_stuff')
    relative_icon = flow.Relative('some_stuff/sub_stuff').ui(label='This is not a child !', icon='asset')

    select = flow.Child(SelectRelative)
    _dyn_relative_oid = flow.Param('some_stuff/sub_stuff')
    dyn_relative = DynRelativeExample()

#---- Connection Example

class Option1(Option):

    ICON = 'maya'

    doc = Label('This is probably not a good option...')

class Option2(Option):

    ICON = '3dsmax'

    doc = Label('Is this even en option?')


class ConnectionExample(Object):

    option_maya = Child(Option1)
    option_3ds = Child(Option2)
    current = Connection(Option)
    other = Connection(Option)

class RelationsGroup(Object):

    doc = Label(
        '<HR><H2>'
        'Relation is what binds objects together.<br>'
        'There are several relations types, all but the "Param" one are showcased here.<br>'
        '(See the Value page for examples of the Param relations)'
        '</H2>'
    )

    child_doc = Label(
        '<HR><H3>'
        'The <u>Child</u> relation let you add an Object under another one:'

    )
    my_child = Child(MyChild)

    parent_doc = Label(
        '<HR><H3>'
        'The <u>Parent</u> relation gives a reference to an Object\'s (grand)parent.<br>'
        'The GUI will show a "goto..." button for each Parent relation of the current '
        'page Object.'
        '</H3>'
        '(Open "Parent Example" and enter "My Child" to see the buttons)'
    )
    parent_example = Child(ParentExample)

    relative_doc = Label(
        '<HR><H3>'
        'The <u>Relative</u> relation gives a reference to an arbitrary Object.<br>'
        'The GUI will show it as a Child even though it\'s not.<br>'
        'Open the "relative_example" in a new page to test it:'
        '</H3>'
    )
    relative_example = Child(RelativeExample)

    conn_doc = Label(
        '<HR><H3>'
        'The <u>Connection</u> relation lets the user specify the related Object.<br>'
        'In this example you can drag and drop one the the two options to the "Current" relation.'
    )
    connection_example = Child(ConnectionExample)

    label_doc = Label(
        '<HR><H3>'
        'Some relation are just GUI sugar, like this "Label" for example.<br>'
        'You can also add simple separators like this one:'
    )
    sep = Separator()

    group_doc = Label(
        '<HR><H3>'
        'You can group your relations with the group context by using "with kabaret.flow.group(&lt;name&gt;):".<br>'
        'Children are grouped together'
    )
    with group('foo'):

        bar = Child(MyChild)
        baz = Child(MyChild)

        with group('yolo'):
            wip = Label(
                '<HR><H3>'
                'Nested groups feature in progress...'
            )

    sep2 = Separator()



import os

from qtpy import QtCore, QtGui, QtWidgets

from .. import Style


class GrayStyle(Style):
    """
    You can customize this style by modifying QSettings() (colors/*)
    """

    def __init__(self, name=None):
        super(GrayStyle, self).__init__(name or 'Gray')

    def apply(self, widget=None):
        widget = widget or QtWidgets.QApplication.instance()

        app = QtWidgets.QApplication.instance()

        # --- Stuff we could not deal with only using css:

        self.set_property('alternate_child_color', False)

        # --- Change palette only for app wide apply:

        if widget is app:
            widget = widget or QtWidgets.QApplication.instance()

            settings = QtCore.QSettings()
            settings.beginGroup('colors')

            # setup the palette
            palette = QtWidgets.QApplication.palette()
            # A color to indicate a selected item or the current item. By default, the highlight color is Qt.darkBlue.
            palette.setColor(QtGui.QPalette.Highlight, settings.value('highlight', QtGui.QColor("#ad4e5c")))
            palette.setColor(QtGui.QPalette.HighlightedText, settings.value('highlighted_text', QtGui.QColor("#42314a")))
            palette.setColor(QtGui.QPalette.WindowText, settings.value('window_text', QtGui.QColor("#b9c2c8")))
            palette.setColor(QtGui.QPalette.Window, settings.value('window', QtGui.QColor("#3e4041")))
            palette.setColor(QtGui.QPalette.Text, settings.value('text', QtGui.QColor("#a7b0b4")))
            palette.setColor(QtGui.QPalette.Base, settings.value('base', QtGui.QColor("#2b2b2b")))
            palette.setColor(QtGui.QPalette.Dark, settings.value('dark', QtGui.QColor("#22222b")))
            palette.setColor(QtGui.QPalette.Light, settings.value('light', QtGui.QColor("#ffe6ee")))
            palette.setColor(QtGui.QPalette.Midlight, settings.value('midlight', QtGui.QColor("#911f36")))
            palette.setColor(QtGui.QPalette.Mid, settings.value('mid', QtGui.QColor("#4b1b1f")))
            palette.setColor(QtGui.QPalette.Button, settings.value('button', QtGui.QColor("#4c5052")))
            palette.setColor(QtGui.QPalette.ButtonText, settings.value('button_text', QtGui.QColor("#a9b7c6")))

            settings.endGroup()

            widget.setPalette(palette)

        # --- Load and apply the css
        this_folder = os.path.dirname(__file__)
        css_file = os.path.join(this_folder, 'gray_style.css')
        with open(css_file, 'r') as r:
            self.apply_css(widget, r.read())






from .interface import Editor_Interface

from qtpy import QtWidgets, QtCore


class PercentValueEditor(QtWidgets.QProgressBar, Editor_Interface):
    '''
    This editor display a progression

    Editor Type Names:
        percent, progression

    Options:
        minimum: float (0)
        maximum: float (100)
        display_text: bool (True) - Display also value in text or not
    '''

    @classmethod
    def can_edit(cls, editor_type_name):
        '''
        Must be implemented to return True if the given editor_type_name
        matches this editor.
        '''
        return editor_type_name in ('percent', 'progress', )

    def __init__(self, parent, options):
        QtWidgets.QProgressBar.__init__(self, parent)
        Editor_Interface.__init__(self, parent)
        self.apply_options(options)

    def set_editable(self, b):
        '''
        Must be implemented to prevent editing if b is False.
        Visual cue show also be given to the user.
        '''
        self.setEnabled(b)

    def apply_options(self, options):
        '''
        Must be implemented to configure the editor as
        described by the options dict.
        '''
        self.setMinimum(options.get('minimum', 0))
        self.setMaximum(options.get('maximum', 100))
        self.setTextVisible(options.get('display_text', True))

    def update(self):
        '''
        Must be implemnented to show the value returned by self.fetch_value()
        Your code should call self._on_updated() at the end.
        '''
        value = float(self.fetch_value())
        self.setValue(value)
        self._on_updated()

    def get_edited_value(self):
        '''
        Must be implemented to return the value currently displayed.
        '''
        return self.value()

    def _show_edited(self):
        '''
        Must be implemented to show that the displayed value
        needs to be applied.
        '''
        # This will not happend, the changes are directly applied.
        self.setProperty('edited', True)
        self.setProperty('applying', False)
        self.style().polish(self)

    def _show_applied(self):
        '''
        Must be implemented to show that the displayed value
        as been saved.
        In a clean scenario, applying edits will trigger an update()
        and this state should disapear.
        If you are using the Editor without this kind of round trip,
        you can call update here.
        '''
        self.setProperty('applying', True)
        self.setText('Applying...')

    def _show_clean(self):
        '''
        Must be implemented to show that the displayed value is
        up to date.
        '''
        self.setProperty('edited', False)
        self.setProperty('applying', False)
        self.style().polish(self)

    def _show_error(self, error_message):
        '''
        Must be implemented to show that the given error occured.
        '''
        self.setProperty('error', True)
        self.style().polish(self)
        self.setToolTip('!!!\nERROR: %s' % (error_message,))


import os
import pluggy

plugin = pluggy.HookimplMarker("kabaret.plugin")
hookspec = pluggy.HookspecMarker("kabaret.plugin")

class PluginSpec:

    @hookspec
    def install_actors(self, session):
        '''
        Install some Actors in the given session.
        '''

    @hookspec
    def install_views(self, session):
        '''
        Install some views in the given session.
        '''

    @hookspec
    def install_resources(self, session):
        '''
        Install/Activate some resources.
        '''

    @hookspec
    def install_editors(self, session):
        '''
        Install some editor factories.
        '''
    
    @hookspec
    def get_project_types(self, session):
        '''
        Must return a list of `kabaret.flow.Object` 
        subclasses intended to be used as 
        project type.
        '''

    @hookspec
    def get_flow_objects(self, session):
        '''
        Must return a list of `kabaret.flow.Object`
        subclasses intended to be used as
        project building blocks (aka, flow lib).
        '''
        

class PluginManager(object):
    """
    Usage:
    ```
    pm = PluginManager()
    pm.register(some_plugin_function)
    pm.register(some_plugin_module)
    pm.register(PluginClass) 
    pm.register(plugin_class_instance)
    pm.load_plugins() # registers plugins from installed packages

    ```
    """
    BLOCK_LIST_ENV_NAME = 'KABARET_BLOCKED_PLUGINS'

    def __init__(self, session):
        super(PluginManager, self).__init__()
        self.session = session
        self._manager = pluggy.PluginManager("kabaret.plugin")
        self._manager.add_hookspecs(PluginSpec)
        self._manager.load_setuptools_entrypoints(group="kabaret.plugin", name=None)
        self.apply_block_list()
        self._hook = self._manager.hook

    def register(self, plugin, name=None):
        """ Register a plugin and return its canonical name or ``None`` if the name
        is blocked from registering.  Raise a :py:class:`ValueError` if the plugin
        is already registered. """
        return self._manager.register(plugin, name)

    def apply_block_list(self, *names_to_block):
        from_env = os.environ.get(self.BLOCK_LIST_ENV_NAME, '')
        for name in list(names_to_block)+from_env.split():
            name = name.strip()
            self._manager.set_blocked(name)
            self.session.log_info('Plugin blocked: {}'.format(name))

    def install_actors(self, session):
        self._hook.install_actors(session=session)

    def install_views(self, session):
        self._hook.install_views(session=session)

    def install_resources(self, session):
        self._hook.install_resources(session=session)

    def install_editors(self, session):
        self._hook.install_editors(session=session)







raise Exception("OBSOLETE")

# from ..editors import EditorFactory

# from ..editors.python_value import PythonValueEditor


# _FACTORY = None


# class FlowPythonValueEditor(PythonValueEditor):

#     def fetch_value(self):
#         '''
#         Must be implemented to return the value to show.
#         '''
#         return self.session.cmds.Flow.get_value(self.value_id)

#     def apply_edited_value(self):
#         '''
#         Must be implemented to apply the value returned 
#         by self.get_edited_value()
#         '''
#         self.session.cmds.Flow.set_value(
#             self.value_id, self.get_edited_value()
#         )


# class FlowEditorFactory(EditorFactory):

#     def __init__(self, session):
#         super(FlowEditorFactory, self).__init__()
#         self.session = session

#     def create(self, value_id, editor_type_name, options, parent):
#         editor = super(FlowEditorFactory, self).create(
#             value_id, editor_type_name, options, parent
#         )
#         editor.session = self.session
#         return editor

# def flow_editor_factory(session):
#     global _FACTORY
#     if _FACTORY is None:
#         _FACTORY = FlowEditorFactory(session)
#         _FACTORY.register_editor_type(FlowPythonValueEditor)
#     else:
#         # we never know (which is a desing flaw, I know :p)
#         _FACTORY.session = session 
#     return _FACTORY


import six
import redis
import os

from .._actor import Actor, Cmd, Cmds


class ClusterCmds(Cmds):

    pass


@ClusterCmds.cmd
class Connect(Cmd):

    def _decode(self, host, port, cluster_name, db_index, password=None):
        self.host = host
        self.port = port
        self.cluster_name = cluster_name
        self.db_index = db_index
        self.password = password

    def _execute(self):
        actor = self.actor()
        actor.log('Connecting to %s port:%r, index:%r' %
                  (self.host, self.port, self.db_index))

        actor.host = self.host
        actor.port = self.port
        actor.db_index = self.db_index
        actor.password = self.password

        # We add those to environment so that subprocesses' session can use connect_from_env():
        os.environ.update(dict(
            KABARET_HOST=self.host,
            KABARET_PORT=self.port,
            KABARET_CLUSTER_NAME=self.cluster_name,
            KABARET_DB=self.db_index,
            KABARET_PASS=self.password or '',
        ))

        # NOTE: decode_responses is needed with python 3 who wont accept bytes as string.
        actor.set_db(
            redis.StrictRedis(
                self.host, self.port, self.db_index, self.password,
                decode_responses=True,
            ),
            self.cluster_name
        )


@ClusterCmds.cmd
class Connect_From_Env(Cmd):
    '''
    Will connect using the env var values.
    '''

    def _decode(self):
        pass

    def _execute(self):
        actor = self.actor()
        actor.log('Connecting From Env...')
        try:
            host = os.environ['KABARET_HOST']
            port = os.environ['KABARET_PORT']
            cluster_name = os.environ['KABARET_CLUSTER_NAME']
            db_index = os.environ['KABARET_DB']
            password = os.environ.get('KABARET_PASS', None) or None
        except (KeyError, ValueError) as err:
            raise Exception(
                'Could not auto_connect the session (env value error): %s' % (err,))
        else:
            actor.log(' => Env Connection: %s port:%r, index:%r' %
                      (host, port, db_index))

        actor.host = host
        actor.port = port
        actor.db_index = db_index
        actor.password = password

        # NOTE: decode_responses is needed with python 3 who wont accept bytes as string.
        actor.set_db(
            redis.StrictRedis(
                host, port, db_index, password,
                decode_responses=True,
            ),
            cluster_name
        )


@ClusterCmds.cmd
class Get_Connection_Info(Cmd):

    def _decode(self):
        pass

    def _execute(self):
        return self.actor().get_connection_info()


@ClusterCmds.cmd
class Disconnect(Cmd):

    def _decode(self):
        pass

    def _execute(self):
        self.actor().log('Disconnecting')


@ClusterCmds.cmd
class Broadcast(Cmd):

    def _decode(self, *words):
        self.words = words

    def _execute(self):
        self.actor().broadcast(*self.words)


@ClusterCmds.cmd
class Publish(Cmd):

    def _decode(self, **channels_messages):
        self.channels_messages = channels_messages

    def _execute(self):
        self.actor().publish(**self.channels_messages)


@ClusterCmds.cmd
class Get_Cluster_Name(Cmd):

    def _decode(self):
        pass

    def _execute(self):
        return self.actor().get_cluster_name()


class Cluster(Actor):

    def __init__(self, session):
        super(Cluster, self).__init__(session)

        self.KABARET_DB_INDEX = 3
        self.db = None
        self.pubsub = None

        self.host = None
        self.port = None
        self.db_index = None
        self.password = None
        self.cluster_name = None

        self._pubsubs = []

    def get_connection_info(self):
        '''
        Returns the info used to connect.
        '''
        return self.host, self.port, self.cluster_name, self.db_index

    def _create_cmds(self):
        return ClusterCmds(self)

    def _reg_pubsub(self, pubsub):
        self._pubsubs.append(pubsub)

    def _unreg_pubsub(self, pubsub, pattern):
        self._pubsubs.remove(pubsub)
        if pattern:
            pubsub.punsubscribe()
        else:
            pubsub.unsubscribe()

    def _read_pubsub(self):
        '''
        Called periodically by the session since self.set_db() called.
        '''
        # nb: we copy self._pubsubs so that handlers can unsubscribe
        # their channel (and thus modify self._pubsubs)

        if 0:
            # This was nice but it fetches only one message per pubsub.
            # If there's many messages, it will take many call to consume
            # every message :/
            for pubsub in list(self._pubsubs):
                unhandled_message = pubsub.get_message(
                    timeout=0   # dont wait for messages
                )
                if unhandled_message is not None:
                    self.on_unhandled_message(unhandled_message)

        else:
            # This is better since we consume messages until there's no
            # more.
            # Maybe we should make the maximum configurable ?
            #
            # NB: this code is derived from redis.client.PubSub.get_message()
            # It might be needed to update it if redis.client changes a lot
            # in further versions.
            for pubsub in list(self._pubsubs):
                max_iterations = 20
                curr = 0
                while curr < max_iterations:
                    response = pubsub.parse_response(block=False, timeout=0)
                    if response:
                        unhandled_message = pubsub.handle_message(
                            response, ignore_subscribe_messages=True
                        )
                        if unhandled_message:
                            self.on_unhandled_message(unhandled_message)
                    else:
                        curr = max_iterations

    def channels_subscribe(self, **channels_callbacks):
        '''
        Returns a a callable used to unsuscribe all those subscriptions.
        '''
        cluster_channels_callbacks = {}
        for ch, cb in channels_callbacks.items():
            channel = '%s:%s' % (self.cluster_name, ch)
            cluster_channels_callbacks[channel] = cb

        pubsub = self.db.pubsub(ignore_subscribe_messages=True)
        pubsub.subscribe(**cluster_channels_callbacks)
        self._reg_pubsub(pubsub)
        return lambda p=pubsub: self._unreg_pubsub(p, pattern=False)

    def patterns_subscribe(self, **patterns_callbacks):
        '''
        Returns a a callable used to unsuscribe all those subscriptions.
        '''
        cluster_patterns_callbacks = {}
        for pt, cb in patterns_callbacks.items():
            pattern = '%s:%s' % (self.cluster_name, pt)
            cluster_patterns_callbacks[pattern] = cb

        pubsub = self.db.pubsub(ignore_subscribe_messages=True)
        pubsub.psubscribe(**cluster_patterns_callbacks)
        self._reg_pubsub(pubsub)
        return lambda p=pubsub: self._unreg_pubsub(p, pattern=True)

    def broadcast(self, *words):
        message = ' '.join([str(i) for i in words])
        self.publish(broadcast=message)

    def publish(self, **channels_messages):
        for channel, message in six.iteritems(channels_messages):
            cluster_channel = '%s:%s' % (self.cluster_name, channel)
            self.db.publish(cluster_channel, message)

    def on_unhandled_message(self, message):
        self.log("Unhandled Message:", repr(message))

    def on_broadcast_message(self, message):
        # message is a dict w/ pattern, type, channel, data
        self.log("[Broadcast Message]", repr(message['data']))

        if 0:
            channels = self.db.pubsub_channels()
            self.log(
                "- Channels:",
                ', '.join(
                    ['%s=%i' % (c, i)
                     for c, i in self.db.pubsub_numsub(*channels)]
                )
            )
            self.log("- Patterns in use:", self.db.pubsub_numpat())

    def get_cluster_name(self):
        return self.cluster_name

    def get_db(self):
        if self.db is None:
            raise Exception('Cannot access the database without connection!')
        return self.db

    def set_db(self, db, cluster_name):
        if self.db is not None:
            raise Exception("The Database is already set !")

        try:
            db.ping()
        except redis.exceptions.ConnectionError as err:
            raise ValueError(
                str(err) + '\n\n'
                '=> Ask your administrator to check the database.'
            )

        self.db = db
        self.db.client_setname(self.session().session_uid())

        self.cluster_name = cluster_name
        self.log('Connected to Cluster %r' % (self.cluster_name,))

        self.session().add_ticked(self._read_pubsub)
        self.session()._on_cluster_connected()

    def on_session_connected(self):
        self.broadcast_unsubscribe = self.channels_subscribe(
            broadcast=self.on_broadcast_message)
        self.broadcast('Cluster joined by', self.session().session_uid())

    def die(self):
        self.broadcast('Cluster left by', self.session().session_uid())
        self.broadcast_unsubscribe()





from qtpy import QtCore, QtWidgets


class FlowLayout(QtWidgets.QLayout):
    def __init__(self, parent=None, margin=0, spacing=0):
        super(FlowLayout, self).__init__(parent)
 
        if parent is not None:
            self.setMargin(margin)
 
        self.setSpacing(spacing)
 
        self.itemList = []
 
    def __del__(self):
        self.clear()

    def clear(self):
        item = self.takeAt(0)
        while item:
            w = item.widget()
            if w:
                w.deleteLater()
            item = self.takeAt(0)

    def addItem(self, item):
        self.itemList.append(item)
 
    def count(self):
        return len(self.itemList)
 
    def itemAt(self, index):
        if index >= len(self.itemList):
            return None

        if index >= 0:
            return self.itemList[index]

        return None
 
    def takeAt(self, index):
        if index >= 0 and index < len(self.itemList):
            return self.itemList.pop(index)
 
        return None
 
    def expandingDirections(self):
        return QtCore.Qt.Orientations(QtCore.Qt.Orientation(0))
 
    def hasHeightForWidth(self):
        return True
 
    def heightForWidth(self, width):
        height = self.doLayout(QtCore.QRect(0, 0, width, 0), True)
        return height
 
    def setGeometry(self, rect):
        super(FlowLayout, self).setGeometry(rect)
        self.doLayout(rect, False)
 
    def sizeHint(self):
        return self.minimumSize()
 
    def minimumSize(self):
        size = QtCore.QSize()
 
        for item in self.itemList:
            size = size.expandedTo(item.minimumSize())
 
        margin = self.contentsMargins()
        size += QtCore.QSize(margin.left()+margin.right(), margin.top()+margin.bottom())
        return size
 
    def doLayout(self, rect, testOnly):
        x = rect.x()
        y = rect.y()
        lineHeight = 0
 
        for item in self.itemList:
            #wid = item.widget()
            spaceX = self.spacing()#+wid.style().layoutSpacing(QtWidgets.QSizePolicy.PushButton, QtWidgets.QSizePolicy.PushButton, QtCore.Qt.Horizontal)
            spaceY = self.spacing()#+wid.style().layoutSpacing(QtWidgets.QSizePolicy.PushButton, QtWidgets.QSizePolicy.PushButton, QtCore.Qt.Vertical)
            nextX = x + item.sizeHint().width() + spaceX
            if nextX - spaceX > rect.right() and lineHeight > 0:
                x = rect.x()
                y = y + lineHeight + spaceY
                nextX = x + item.sizeHint().width() + spaceX
                lineHeight = 0
 
            if not testOnly:
                item.setGeometry(QtCore.QRect(QtCore.QPoint(x, y), item.sizeHint()))
 
            x = nextX
            lineHeight = max(lineHeight, item.sizeHint().height())
 
        return y + lineHeight - rect.y()



import os
import logging
import subprocess
from collections import OrderedDict

from kabaret.flow import (
    values,
    Object, ThumbnailInfo, Map, Action, ConnectAction,
    ChoiceValueSetAction, ChoiceValueSelectAction,
    Child, Parent, Computed, Connection,
    Param, IntParam, BoolParam, HashParam
)

#-------------------------------- ACTIONS


class ExploreAction(Action):

    SHOW_IN_PARENT_INLINE = False
    ICON = 'explorer'


class TextEditActionOptions(Object):

    those = Param()
    are = Param()
    grouped = Param()
    action = Param()
    params = Param()


class TextEditAction(Action):

    ICON = 'notepad'

    options = Child(TextEditActionOptions)


class MayaEditAction(Action):

    ICON = 'maya'

    _parent = Parent()

    def get_context_oid(self):
        return self._parent.get_maya_context_oid()

    def needs_dialog(self):
        return False

    def run(self, button):
        maya = 'D:/Autodesk/Maya2015/bin/maya.exe'
        env = os.environ.copy()

        # The idea is to have the code configuring maya's embedded kabaret
        # somewhere inside you studio.
        # Here we have:
        #   dev_studio.flow.demo_project.maya_startup
        # It contains a ultra simple userSetup.py that calls something like
        #   dev_studio.flow.demo_project.maya_startup.install()
        # which in turn does all you want.

        # The best way to have you studio inside maya is
        # a shared install (It avoid installing it inside maya on
        # every workstation)
        # But beware of your .pyc files ! If your python is not the
        # same as Mayapy you should have a separated installation
        # for import made from inside maya.
        #
        # An easy way to deal with this is to set a studio env var like
        #   os.environ["MY_STUDIO_PYTHONDIR"] = "path/to/maya_pyton_libs"
        # And use this in maya's userSetup.py to extend sys.path and import
        # your studio.

        # For this demo, I assume the pyc are compatible (which is BAAAAA
        # AAAAAD !) and will use the current kabaret and dev_studio path:
        import kabaret
        env['KABARET_INSTALL'] = os.path.dirname(
            os.path.dirname(os.path.abspath(kabaret.__file__))
        )

        import dev_studio
        env['DEV_STUDIO_INSTALL'] = os.path.dirname(
            os.path.dirname(os.path.abspath(dev_studio.__file__))
        )

        # # In order to start a Kabaret Session for this project, maya code
        # # will need to know the cluster connection arguments.
        # # Most of the time, you'd better get them from a function in your
        # # studio. But for this demo we use command line args so I'll just
        # # use env to pass the info to maya:
        # host, port, cluster_name = self.root().session().cmds.Cluster.get_connection_info()
        # env['KABARET_HOST'] = host
        # env['KABARET_PORT'] = port
        # env['KABARET_CLUSTER_NAME'] = cluster_name

        # The Context view will load the oid defined in this env var:
        env['KABARET_CONTEXT_OID'] = str(self.get_context_oid())

        # Use PYTHONPATH to make maya exec our userSetup.py at startup:
        from . import maya_startup
        userSetup_dir = os.path.dirname(maya_startup.__file__)
        PYTHONPATH = os.environ.get('PYTHONPATH', '')
        env['PYTHONPATH'] = userSetup_dir + ';' + PYTHONPATH

        # Obviously the action should open a file :p
        # Adding a scene = Parent() relation would let you do something like:
        #   scene_filename = self.scene.get_filename()
        #   command_args = [maya, scene_filename]
        # But I dont deal with filename in this demo (not yet at least ^.^)
        # So:
        command_args = [maya]

        # Here you go !
        popen = subprocess.Popen(command_args, env=env)


class SetWAIT(ChoiceValueSetAction):
    ICON = ('icons.status', 'WAIT')
    VALUE_TO_SET = 'WAIT'


class SetWIP(ChoiceValueSetAction):
    ICON = ('icons.status', 'WIP')
    VALUE_TO_SET = 'WIP'


class SetRVW(ChoiceValueSetAction):
    ICON = ('icons.status', 'RVW')
    VALUE_TO_SET = 'RVW'


class SetRTK(ChoiceValueSetAction):
    ICON = ('icons.status', 'RTK')
    VALUE_TO_SET = 'RTK'


class SetDONE(ChoiceValueSetAction):
    ICON = ('icons.status', 'DONE')
    VALUE_TO_SET = 'DONE'


class StatusValue(values.ChoiceValue):

    STRICT_CHOICES = True
    CHOICES = ['NYS', 'WAIT', 'WIP', 'RVW', 'RTK', 'DONE', 'OOP']

    wait = Child(SetWAIT)
    wip = Child(SetWIP)
    rvw = Child(SetRVW)
    rtk = Child(SetRTK)
    done = Child(SetDONE)
    other = Child(ChoiceValueSelectAction)


class Status(Param):

    def __init__(self, default_value='NYS'):
        super(Status, self).__init__(default_value, StatusValue)


# ----------------------------------------- ASSET

class Asset(Object):

    ICON = 'asset'

    some_param = Param('some value...')


class CreateAssetAction(Action):

    _map = Parent()

    asset_preffix = Param('MyAsset')

    def get_buttons(self):
        return ['Create']

    def run(self, button):
        preffix = self.asset_preffix.get()
        asset_name = '%s_%03i' % (preffix, len(self._map),)
        self._map.add(asset_name)
        self._map.touch()


class Assets(Map):

    def mapped_type(cls):
        return Asset

    create_asset = Child(CreateAssetAction)

# ----------------------------------------- BANK


class Bank(Object):

    ICON = 'bank'

    assets = Child(Assets)


class CreateBankAction(Action):

    _map = Parent()

    bank_name = Param('Bank')

    def get_buttons(self):
        return ['Create']

    def run(self, button):
        self._map.add(self.bank_name.get())
        self._map.touch()


class Banks(Map):

    ICON = 'bank'

    # _items = HashParam(
    #     OrderedDict([
    #         ('Bank', ' my_studio.flow.test_project.Bank'),
    #     ])
    # )

    create_bank = Child(CreateBankAction)

    @classmethod
    def mapped_type(cls):
        return Bank

# --------------------------------------- SHOT


class Montage(Object):

    index = IntParam(0)
    first = IntParam(1).watched()
    fade_in = IntParam(0)
    last = IntParam(100).watched()
    fade_out = IntParam(0)
    nb_frames = Computed()

    def child_value_changed(self, child_value):
        if child_value in (self.first, self.last):
            self.nb_frames.touch()

    def compute_child_value(self, child_value):
        if child_value == self.nb_frames:
            self.nb_frames.set(
                self.last.get() - self.first.get() + 1
            )


class CastItem(Object):

    asset = Connection(Asset)
    index = IntParam(0)

    def asset_name(self):
        asset = self.asset.get()
        if asset is None:
            return None
        return asset.name()


class TestMultipleConnectActions(ConnectAction):

    casting = Parent()

    def accept_label(self, objects, urls):
        return "This does nothing"

    def run(self, objects, urls):
        logging.getLogger('kabaret.demo').info('Doing nothing with %r and %r' % (objects, urls))


class CastingAssetConnectAction(ConnectAction):

    casting = Parent()

    def accept_label(self, objects, urls):
        logging.getLogger('kabaret.demo').info("DROP!")
        if urls:
            return None
        for o in objects:
            if not isinstance(o, Asset):
                return None
        return "Add %i Assets to the Casting" % (len(objects,))

    def run(self, objects, urls):
        for asset in objects:
            ci = self.casting.add_asset(asset)
        self.casting.touch()


class Casting(Map):

    _drop_asset = Child(CastingAssetConnectAction)
    drop_something = Child(TestMultipleConnectActions)

    @classmethod
    def mapped_type(cls):
        return CastItem

    def add_asset(self, asset):
        used_indexes = set()
        for ci in self.mapped_items():
            c_asset = ci.asset.get()
            if c_asset is asset:
                used_indexes.add(ci.index.get())
        if not used_indexes:
            index = 1
        else:
            index = None
            for i in range(min(used_indexes), max(used_indexes)):
                if i not in used_indexes:
                    index = i
            if index is None:
                index = max(used_indexes) + 1
        name = asset.name() + ('_id%03i' % index)
        ci = self.add(name)
        ci.asset.set(asset)
        ci.index.set(index)

        return ci


class ShotObject(Object):

    def summary(self):
        return (
            '<font color=#440000><b>'
            '| File does not exists | Not Yet Reversionned |'
            '</b></font>'
        )


class AddToReversionAction(Action):
    SHOW_IN_PARENT_INLINE = False


class ProtectedAction(Action):
    SHOW_IN_PARENT_INLINE = False


class FileHistoryMap(Map):
    pass


class PublishAction(Action):
    ICON = ('icons.gui', 'share-option')


class InitAction(Action):
    pass


class DummyMayaAction(Action):

    pass


class MayaContextToolsGroup(Object):

    sub_tool01 = Child(DummyMayaAction)
    sub_tool02 = Child(DummyMayaAction)
    sub_tool03 = Child(DummyMayaAction)


class MayaContextTools(Object):

    group01 = Child(MayaContextToolsGroup)
    tool01 = Child(DummyMayaAction)
    tool02 = Child(DummyMayaAction)
    group02 = Child(MayaContextToolsGroup)
    tool03 = Child(DummyMayaAction)


class ShotTask(ShotObject):

    filename = Computed().ui(group='file')
    exists = Computed().ui(editor='bool').ui(group='file')
    explore = Child(ExploreAction).ui(group='file')

    is_versionned = BoolParam().ui(group='cvs')
    activate_versionning = Child(AddToReversionAction).ui(group='cvs')
    locked_by = Computed().ui(group='cvs')
    steal_work = Child(ProtectedAction).ui(group='cvs')
    trash_work = Child(ProtectedAction).ui(group='cvs')
    history = Child(FileHistoryMap).ui(group='cvs')

    publish = Child(PublishAction).ui(group='cvs')

    edit = Child(TextEditAction).ui(group='work')
    initialize = Child(InitAction).ui(group='work')

    open = Child(MayaEditAction).ui(group='work')
    _maya_context_tools = Child(MayaContextTools)

    @classmethod
    def get_source_display(self, oid):
        shot_oid, task = oid.rsplit('/', 1)
        return (
            Shot.get_source_display(shot_oid)
            +
            ' -> ' + task
        )

    def compute_child_value(self, child_value):
        child_value.set({
            self.filename: 'path/to/filename.ext',
            self.exists: False,
            self.is_versionned: False,
            self.locked_by: '-?-',
        }[child_value])

    def get_maya_context_oid(self):
        return self._maya_context_tools.oid()


class Anim(ShotObject):

    ICON = 'maya'


class Lighting(ShotObject):

    ICON = 'alembic'


class Comp(ShotObject):

    ICON = 'natron'


class SomeAction(Action):

    ICON = 'shot'

    def needs_dialog(self):
        return False

    def run(self, button):
        logging.getLogger('kabaret.demo').info('ACTION: ' + self.oid())


class DemoShotThumbnail(ThumbnailInfo):
    '''
    This class generates different thumbnail for demo purpose
    '''

    _RESOURCE = 'shot'
    _IMAGE = 'E:/tmp/Computer-Guy-Facepalm.jpg'

    _shot = Parent()

    def shot_num(self):
        return int(self._shot.name()[-1])

    def is_resource(self):
        return not self.shot_num() % 3

    def get_resource(self):
        return self._RESOURCE

    def is_image(self):
        return not (self.shot_num() - 1) % 3

    def is_sequence(self):
        return not (self.shot_num() - 2) % 3

    def get_label(self):
        return None

    def get_path(self):
        if self.is_sequence():
            p = 'E:/tmp/TEST_SEQ/HD_1420x594/M2_S2210_P0100-anim_blasts.%04i.jpg'
            p = 'E:/tmp/TEST_SEQ/LD_355x149/seq_ld.%04i.jpg'
            p = 'M:/KABARET_PROJECTS/MINUS/SYSTEM/RVS/REPO/films/M2/S2210/P0100/blast/LastBlast=M2_S2210_P0100-Animation-Blast/M2_S2210_P0100-anim_blasts.%04i.png'
            return p
        return self._IMAGE

    def get_default_height(self):
        return 200

    def get_first_last_fps(self):
        if self.is_sequence():
            return 1001, 1179, 24
        return None, None, None


class Shot(Object):

    ICON = 'shot'

    _thumbnail = Child(DemoShotThumbnail)

    status = Status().watched().ui(group='Prod')
    montage = Child(Montage).ui(group='Prod')

    film = Parent(2)

    casting = Child(Casting).ui(group='Prod')
    animatic = Child(ShotTask).ui(group='Tasks')
    sound = Child(ShotTask).ui(group='Tasks')
    anim = Child(ShotTask).ui(group='Tasks')
    lighting = Child(ShotTask).ui(group='Tasks')
    comp = Child(ShotTask).ui(group='Tasks')

    @classmethod
    def get_source_display(cls, oid):
        names = oid.split('/')
        return ' '.join(names[-3::2])

    def child_value_changed(self, child_value):
        if child_value is self.status:
            self.touch()

    def summary(self):
        return '%s (Anim, Lighting)' % (self.status.get(),)

    def get_thumbnail_object(self):
        '''
        Implementing this method let the GUI know how to
        display images for this object
        '''
        return self._thumbnail

# --------------------------------------- FILM


class Add10ShotsAction(Action):

    ICON = 'sequence'

    _shots = Parent()

    def run(self, button):
        start = len(self._shots) + 1
        for i in range(100):
            self._shots.add('P%03i' % (i + start))

        self._shots.touch()


class Shots(Map):

    ICON = 'shot'

    add_10_shots = Child(Add10ShotsAction)

    @classmethod
    def mapped_type(cls):
        return Shot

    def columns(self):
        return ['Name', 'Status']

    def _fill_row_cells(self, row, item):
        '''
        Subclasses must override this to fill value for each
        column returned by columns()
        '''
        row['Name'] = item.name()
        row['Status'] = item.status.get()

    def _fill_row_style(self, style, item, row):
        style['Status_icon'] = ('icons.status', row['Status'])


class Film(Object):

    shots = Child(Shots).ui(default_height=500)


class CreateFilmAction(Action):

    _map = Parent()

    film_name = Param('EP')

    def get_buttons(self):
        return ['Create']

    def run(self, button):
        i = len(self._map) + 1
        film = self._map.add('%s%03i' % (self.film_name.get(), i))
        self._map.touch()
        return self.get_result(goto=film.oid())


class Films(Map):

    create_film = Child(CreateFilmAction)

    @classmethod
    def mapped_type(cls):
        return Film

# ------------------------------------------------- PROJECT


class Launcher(Action):
    pass


class Launchers(Object):

    maya = Child(Launcher)
    max = Child(Launcher)
    guerilla = Child(Launcher)

    nuke = Child(Launcher)
    natron = Child(Launcher)


class ImageSettings(Object):

    frame_rate = Param(25)
    width = Param(2840)
    height = Param(1188)
    padding = Param(4)
    par = Param(1.0)


class MayaSettings(Object):

    project_path = Param('path/to/maya_project')
    user_away_path = Param('path/to/AWAY/Users/me')


class Settings(Object):

    project = Parent()

    image = Child(ImageSettings)
    Maya = Child(MayaSettings)


class CreateProjectFoldersAction(Action):

    def get_buttons(self):
        return ['Create']

    def run(self, button):
        pass


class ReversionAdmin(Object):

    repo_path = Param('path/to/cvs_repo')
    repo_exists = Param(True).ui(editor='bool')

    work_path = Param('path/to/work_dir')
    work_exists = Param(True).ui(editor='bool')

    mount_script = Param('\n\ntest...\n\n').ui(editor='textarea')


class ProjectMonkeyPatch(Object):

    filename = Param('path/to/project_monkey_patch.py')
    explore = Child(ExploreAction)
    edit = Child(TextEditAction)


class Admin(Object):

    project = Parent()

    create_folders = Child(CreateProjectFoldersAction)
    reversion = Child(ReversionAdmin)
    project_monkey_patch = Child(ProjectMonkeyPatch)


class DemoProject(Object):

    open = Child(Launchers)
    banks = Child(Banks)
    films = Child(Films)

    settings = Child(Settings)
    admin = Child(Admin)



import six
import logging
from qtpy import QtWidgets, QtGui, QtCore

from kabaret.app import resources

from .. import event_filters
from .flow_field import FlowField, FormField, ObjectSummary, ObjectActionMenuManager, ObjectActionsMenu


class _MapField(FormField):

    def __init__(self, parent, map_widget):
        self.map_widget = map_widget
        super(_MapField, self).__init__(parent)
        self.setFlags(QtCore.Qt.NoItemFlags)

    def build(self):
        self.setItemWidget(1, self.map_widget)

    def on_touch_event(self, oid):
        pass


class MapFieldTreeItem(QtWidgets.QTreeWidgetItem):

    def __init__(self, parent, oid, row):
        super(MapFieldTreeItem, self).__init__(parent)
        self.oid = oid
        self.update_row(row)

    def get_activate_oid(self, col):
        column_name = self.treeWidget().column_name(col)
        k = column_name+'_activate_oid'
        style = self._row.get('_style', {})
        try:
            return style[k]
        except KeyError:
            return style.get('activate_oid', None)

    def update_row(self, row):
        tree = self.treeWidget()
        columns = tree.columns
        self._row = row

        style = row.get('_style', {})
        icon = style.get('icon')
        if icon:
            if isinstance(icon, six.string_types):
                icon_ref = ('icons.flow', icon)
            else:
                icon_ref = icon
            try:
                icon = resources.get_icon(icon_ref, tree)
            except resources.NotFoundError as err:
                logging.getLogger('kabaret.ui').debug("WARNING: RESOURCE NOT FOUND: %r" % (err,))
                self.setIcon(0, QtGui.QIcon())
            else:
                self.setIcon(0, icon)

        for name in ('background-color', 'background_color'):
            background = style.get('background-color')
            if background:
                if isinstance(background, six.string_types):
                    brush = QtGui.QBrush(QtGui.QColor(background))
                else:
                    brush = QtGui.QBrush(QtGui.QColor(*background))
                for i in range(self.columnCount()):
                    self.setBackground(i, brush)
                break

        for name in ('foreground-color', 'foreground_color'):
            foreground = style.get(name)
            if foreground:
                if isinstance(foreground, six.string_types):
                    brush = QtGui.QBrush(QtGui.QColor(foreground))
                else:
                    brush = QtGui.QBrush(QtGui.QColor(*foreground))
                for i in range(self.columnCount()):
                    self.setForeground(i, brush)
                break

        for i, col in enumerate(columns):
            icon = style.get(col + '_icon')
            if icon:
                if isinstance(icon, six.string_types):
                    icon_ref = ('icons.flow', icon)
                else:
                    icon_ref = icon
                try:
                    icon = resources.get_icon(icon_ref, tree)
                except resources.NotFoundError as err:
                    logging.getLogger('kabaret.ui').debug("WARNING: RESOURCE NOT FOUND: %r" % (err,))
                    # remove any previously set icon
                    self.setIcon(i, QtGui.QIcon())
                else:
                    self.setIcon(i, icon)

            for suffix in ('_background-color', '_background_color'):
                background = style.get(col + suffix)
                if background:
                    if isinstance(background, six.string_types):
                        brush = QtGui.QBrush(QtGui.QColor(background))
                    else:
                        brush = QtGui.QBrush(QtGui.QColor(*background))
                    self.setBackground(i, brush)
                    break

            for suffix in ('_foreground-color', '_foreground_color'):
                foreground = style.get(col + suffix)
                if foreground:
                    if isinstance(foreground, six.string_types):
                        brush = QtGui.QBrush(QtGui.QColor(foreground))
                    else:
                        brush = QtGui.QBrush(QtGui.QColor(*foreground))
                    self.setForeground(i, brush)
                    break

            self.setText(i, str(row.get(col, '!')))


class MapFieldTree(QtWidgets.QTreeWidget):

    def __init__(self, field, parent):
        super(MapFieldTree, self).__init__(parent)
        self.field = field
        self.session = field.session

        self.setRootIsDecorated(False)
        self.setColumnCount(0)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.setSortingEnabled(True)
        self.setDragEnabled(True)
        self.setDragDropMode(QtWidgets.QAbstractItemView.DragDrop)
        self.setDropIndicatorShown(True)
        self.setProperty("table", True)
        self.columns = []
        self.auto_fit = True
        self.columns_width = None
        self.sort_by = -1
        self.sort_order = _str_to_qt_order('asc')

        self.installEventFilter(self)

    def apply_options(self, ui_config):
        self.sort_by = ui_config.get('sort_by', self.sort_by)
        self.sort_order = _str_to_qt_order(ui_config.get('sort_order', 'asc'))
        self.auto_fit = ui_config.get('auto_fit', self.auto_fit)
        self.columns_width = ui_config.get('columns_width', self.columns_width)
        style = ui_config.get('style', {})
        stylesheet = ""
        if 'header_color' in style:
            stylesheet += \
                'QHeaderView::section {' \
                '   background-color: rgb(%d,%d,%d);' \
                '}' % style.get('header_color')
        if 'background_color' in style:
            stylesheet += \
                'QTreeView[table = "true"] {' \
                '   background-color: rgb(%d,%d,%d);' \
                '}' % style.get('background_color')
        if stylesheet:
            self.setStyleSheet(stylesheet)

    def column_name(self, col):
        return self.columns[col]

    def column_id(self, column_text):
        for i, col in enumerate(self.columns):
            if col.upper() == column_text.upper():
                return i
        return -1

    def set_columns(self, columns):
        self.columns = columns
        self.setHeaderLabels(self.columns)
        self.setColumnCount(len(self.columns))

    def set_items(self, items):
        for oid, row in items:
            MapFieldTreeItem(self, oid, row)

        self.update_column_width()

        if isinstance(self.sort_by, int):
            self.sortItems(self.sort_by, self.sort_order)
        elif isinstance(self.sort_by, str):
            self.sortItems(self.column_id(self.sort_by), self.sort_order)

    def update_column_width(self):
        if self.auto_fit:
            [self.resizeColumnToContents(i) for i, col in enumerate(self.columns)]
        elif self.columns_width:
            for i, width in enumerate(self.columns_width):
                self.setColumnWidth(i, self.width() * width / 100)

    def resizeEvent(self, *args, **kwargs):
        self.update_column_width()
        super(MapFieldTree, self).resizeEvent(*args, **kwargs)

    def mimeData(self, items):
        # print(items)

        mime_data = super(MapFieldTree, self).mimeData(items)
        oids = [item.oid for item in items]
        md = self.session.cmds.Flow.to_mime_data(oids)
        for data_type, data in six.iteritems(md):
            mime_data.setData(data_type, data)

        return mime_data

    def supportedDropActions(self):
        return QtCore.Qt.CopyAction

    def dragEnterEvent(self, event):
        if not event.mouseButtons() & QtCore.Qt.LeftButton:
            event.ignore()
            return
        if self.session.cmds.Flow.can_handle_mime_formats(
                event.mimeData().formats()
        ):
            event.acceptProposedAction()
        else:
            super(MapFieldTree, self).dragEnterEvent(event)

    def dragMoveEvent(self, event):
        if self.session.cmds.Flow.can_handle_mime_formats(
                event.mimeData().formats()
        ):
            event.acceptProposedAction()
        else:
            super(MapFieldTree, self).dragMoveEvent(event)

    def dropMimeData(self, parent, index, mime_data, drop_action):
        md = {}
        for format in mime_data.formats():
            md[format] = mime_data.data(format).data()
        oids, urls = self.session.cmds.Flow.from_mime_data(md)

        if not oids and not urls:
            return False  # let the event propagate up

        connection_targets = self.session.cmds.Flow.get_connection_targets(
            self.field.oid, oids, urls
        )
        if not connection_targets:
            return False  # let the event propagate up

        if len(connection_targets) == 1:
            target_oid, label, icon = connection_targets[0]
            self.session.cmds.Flow.connect(
                target_oid, oids, urls
            )
            return True

        m = QtWidgets.QMenu(self)
        for target_oid, label, icon_name in connection_targets:
            a = m.addAction(label)
            a.triggered.connect(
                lambda
                    toggled=None,
                    target_oid=target_oid,
                    oids=oids,
                    urls=urls: self.field.do_connect(
                    target_oid, oids, urls)
            )
            icon = self.field._get_icon(icon_name)
            if icon is not None:
                a.setIcon(icon)
        m.exec_(QtGui.QCursor.pos())

        return True

    def mousePressEvent(self, e):
        super(MapFieldTree, self).mousePressEvent(e)

        if e.button() == QtCore.Qt.RightButton:
            # NB: We set CustomContextMenu only here to prevent the
            # menu to be updated / shown uppon mouseReleaseEvent wich
            # can hapen far from the item
            self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            self.customContextMenuRequested.emit(e.pos())
            self.setContextMenuPolicy(QtCore.Qt.DefaultContextMenu)

    def mouseMoveEvent(self, e):
        if not e.buttons() & QtCore.Qt.LeftButton:
            e.ignore()
            return
        super(MapFieldTree, self).mouseMoveEvent(e)

    def mouseReleaseEvent(self, e):
        super(MapFieldTree, self).mouseReleaseEvent(e)

    def get_item(self, oid):
        # TODO: see if this is worth optimizing this with a id->item map
        for i in range(self.topLevelItemCount()):
            item = self.topLevelItem(i)
            # print('       ', item.oid)
            if item.oid == oid:
                return item
        return None

    def apply_filter(self, filter):
        it = QtWidgets.QTreeWidgetItemIterator(self)
        col_indexes = range(self.columnCount())
        lfilter = filter.lower()
        if ':' in lfilter:
            col_index, lfilter = lfilter.split(':', 1)
            try:
                col_index = int(col_index)
            except ValueError:
                pass
            else:
                col_indexes = [col_index - 1]

        while it.value():
            item = it.value()
            matched = False
            for c in col_indexes:
                if lfilter in item.text(c).lower():
                    matched = True
                    break
            item.setHidden(not matched)
            it += 1

    def eventFilter(self, source, event):
        if event.type() == QtCore.QEvent.Wheel:
            event.ignore()
            if self.verticalScrollBar().isVisible():
                event.accept()
                try:
                    # Qt4
                    degrees = event.delta() / 8
                except AttributeError:
                    # Qt5
                    degrees = event.angleDelta().y() / 8
                scroll_bar = self.verticalScrollBar()
                scroll_bar.setValue(scroll_bar.value() - scroll_bar.singleStep() * degrees / 15)
                return True
            return False
        return super(MapFieldTree, self).eventFilter(source, event)


class MapFieldFilterLineEdit(QtWidgets.QLineEdit):

    def __init__(self, parent, map_tree):
        super(MapFieldFilterLineEdit, self).__init__(parent)
        self.map_tree = map_tree
        self.textEdited.connect(self._on_edit)

        self.setToolTip(
            'Filter list (not case sensitive)\n'
            'Use "1:blah" to match only first column.'
        )

    def _on_edit(self, text):
        self.map_tree.apply_filter(text)

    def dropEvent(self, e):
        md = e.mimeData()
        if md.hasFormat('kabaret/flow_oid'):
            oid = str(md.data('kabaret/flow_oid'))
            e.acceptProposedAction()
            self.setText(oid[0])
        else:
            super(MapFieldFilterLineEdit, self).dropEvent(e)


def _str_to_qt_order(string):
    return QtCore.Qt.AscendingOrder if string.startswith("a") else QtCore.Qt.DescendingOrder


class MapField(FlowField):

    def __init__(self, parent, session, oid, ui_config=None):
        self._mapped_action_manager = None
        self._mapped_action_menu = None
        self._map_tree = None
        super(MapField, self).__init__(parent, session, oid, ui_config)

    def goto(self):
        self.treeWidget().goto(self.oid)

    def build_children(self):
        tree = self.treeWidget()

        w = QtWidgets.QWidget(tree)
        resizer = event_filters.MouseResizer(w, self.update_height, horizontal=False)
        resizer.propagate_events = False
        w.installEventFilter(resizer)

        lo = QtWidgets.QVBoxLayout()
        lo.setContentsMargins(2, 2, 2, 2)
        lo.setSpacing(0)
        w.setLayout(lo)

        self._map_tree = MapFieldTree(self, None)
        self._map_tree.itemDoubleClicked.connect(
            self._on_mapped_double_clicked
        )
        self._map_tree.customContextMenuRequested.connect(
            self._on_mapped_context_menu
        )
        self._map_tree.installEventFilter(
            event_filters.IgnoreMouseButton(self._map_tree)
        )
        self._map_tree.viewport().installEventFilter(
            event_filters.IgnoreMouseButton(self._map_tree)
        )

        self._map_tree.apply_options(self.ui_config)

        if self.ui_config.get('show_filter', False):
            fle = MapFieldFilterLineEdit(w, self._map_tree)
            fle.setVisible(True)
            fle.setPlaceholderText("Filter...")
            w.layout().addWidget(fle)

        lo.addWidget(self._map_tree)

        default_height = self.ui_config.get('default_height')
        if default_height:
            w.setFixedHeight(default_height)

        self.addChild(_MapField(self, w))
        self.update_content()

    def activated(self, col):
        if not super(MapField, self).activated(col):
            # It it did not open another window:
            if col == 0 or not self.ui_config.get("expandable", True):
                self.goto()
            else:
                if not self.isExpanded():
                    self.ensure_children_built()
                    self.setExpanded(True)
                else:
                    self.setExpanded(False)

    def _inspect_touch_event(self, oid):
        if super(MapField, self)._inspect_touch_event(oid):
            return True
        if self._map_tree:
            item = self._map_tree.get_item(oid)
            if item is not None:
                row = self.session.cmds.Flow.get_mapped_row(self.oid, oid)
                item.update_row(row)
                return True
        # print(':/')
        return False

    def update_content(self):
        if self._map_tree:
            self._map_tree.clear()

            self._map_tree.set_columns(self.session.cmds.Flow.get_mapped_columns(self.oid))
            self._map_tree.set_items(self.session.cmds.Flow.get_mapped_rows(self.oid))
        self._summary.load_summary(self.oid)


    def update_height(self, w):
        self.treeWidget().scheduleDelayedItemsLayout()

    def build(self):
        # self.setTextAlignment(0, QtCore.Qt.AlignCenter|QtCore.Qt.AlignRight)
        self.setFlags(QtCore.Qt.ItemIsEnabled)  # not selectable

        tree = self.treeWidget()

        self._mapped_action_manager = ObjectActionMenuManager(
            self.session, tree.show_action_dialog, 'Flow.map'
        )
        self._mapped_action_menu = QtWidgets.QMenu(tree)

        top_w = QtWidgets.QWidget(tree)
        top_lo = QtWidgets.QHBoxLayout()
        top_lo.setContentsMargins(2, 2, 2, 2)
        top_w.setLayout(top_lo)

        actions_menu = ObjectActionsMenu(
            self.session, top_w, tree.show_action_dialog, 'Flow.inline'
        )
        actions_menu.set_with_submenus(
            self.ui_config.get('action_submenus', False)
        )
        top_lo.addWidget(actions_menu)

        self._summary = ObjectSummary(self.session, top_w)
        self._summary.load_summary(self.oid)
        top_lo.addWidget(self._summary)
        top_lo.addStretch()

        actions_menu.load_actions(self.oid)
        self.setIcon(0, self._get_config_icon())
        self.setText(0, self._label)
        self.setItemWidget(1, top_w)
        self.show_may_have_children()

    def _on_mapped_double_clicked(self, item, col):
        activate_oid = item.get_activate_oid(col)
        if activate_oid is None:
            self.treeWidget().goto(item.oid)
        else:
            self.treeWidget().show_action_dialog(activate_oid)
            
    def _on_mapped_context_menu(self, point):
        if self._map_tree:
            selected = self._map_tree.selectedItems()
            # double check with itemAt bc there's a thing
            # going on when... whatever... trust me -__-"
            item = self._map_tree.itemAt(point)
            if not item:
                return

            try:
                selected[1]
            except IndexError:
                # fall back to single item when only one is selected:
                selected = None

            if not selected:
                got_actions = self._mapped_action_manager.update_oid_menu(
                    item.oid, self._mapped_action_menu,
                    with_submenus=self.ui_config.get('items_action_submenus', False),
                )
            else:
                got_actions = self._mapped_action_manager.update_oids_menu(
                    [i.oid for i in selected], self._mapped_action_menu,
                    with_submenus=self.ui_config.get('items_action_submenus', False),
                )

            if got_actions:
                self._mapped_action_menu.exec_(
                    self._map_tree.viewport().mapToGlobal(point)
                )

    def do_connect(self, target_oid, source_oids, source_urls):
        # print("Connecting", target_oid, source_oids, source_urls)
        self.session.cmds.Flow.connect(
            target_oid, source_oids, source_urls
        )



import six
from qtpy import QtWidgets, QtCore



class EatEnterKeys(QtCore.QObject):

    def eventFilter(self, obj, evt):
        if evt.type() == QtCore.QEvent.KeyPress:
            key = evt.key()
            if key == QtCore.Qt.Key_Enter or key == QtCore.Qt.Key_Return:
                return True  # mark the event as handled
        return False


class IgnoreMouseButton(QtCore.QObject):

    def __init__(self, parent):
        super(IgnoreMouseButton, self).__init__(parent)

        self.buttons = QtCore.Qt.MidButton
        self._events = (
            QtCore.QEvent.MouseButtonPress,
            QtCore.QEvent.MouseMove,
            QtCore.QEvent.MouseButtonRelease,
        )

    def eventFilter(self, obj, e):
        if e.type() in self._events and e.buttons() & self.buttons:
            e.ignore()
            return False # mark handled
        return False

class MouseResizer(QtCore.QObject):
    '''
    Resize the widget with self.button (Defaults to QtCore.Qt.MidButton).
    Optional on_resized(target) is called after each resize.
    
    If aspect_ratio is not None, both horizontal and vertical are considered True.


    '''
    def __init__(self, target, on_resized=None, horizontal=True, vertical=True, aspect_ratio=None):
        super(MouseResizer, self).__init__(target)
        self.propagate_events = True
        
        self._on_resized = on_resized
        self.horizontal = horizontal
        self.vertical = vertical
        self.aspect_ratio = aspect_ratio

        self.target = target
        self.down_pos = None
        self.down_size = None

        self.buttons = QtCore.Qt.MidButton
        self.min_w = 32
        self.min_h = 32

        self._has_resized = None
        self._mns = self.target.minimumSize()
        self._mxs = self.target.maximumSize()
        self._sp = self.target.sizePolicy()

    def _do_on_resized(self):
        if self._on_resized is None:
            return
        self._on_resized(self.target)

    def apply_resize(self, from_width, from_height, x_move, y_move):
        if self.aspect_ratio is not None:
            h = max(
                self.min_h,
                from_height+y_move
            )
            w = int(h*self.aspect_ratio)
            self.target.setFixedSize(w,h)
        else:
            if self.horizontal:
                self.target.setFixedWidth(
                    max(
                        self.min_w,
                        from_width+x_move
                    )
                )
            if self.vertical:
                self.target.setFixedHeight(
                    max(
                        self.min_h,
                        from_height+y_move
                    )
                )
        self._do_on_resized()

    def eventFilter(self, obj, e):
        if e.type() == QtCore.QEvent.MouseButtonPress and e.buttons() & self.buttons:
            self._has_resized = False
            self.down_pos = e.pos()
            self.down_size = self.target.size()

        elif (
            self.down_pos is not None and 
            e.type() == QtCore.QEvent.MouseMove and 
            e.buttons() & self.buttons
        ):
            self._has_resized = True
            o = e.pos() - self.down_pos
            self.apply_resize(
                self.down_size.width(), self.down_size.height(), 
                o.x(), o.y()
            )
            if not self.propagate_events:
                return True
            # we should always return True, but returning None
            # gives a pretty cool side effect on flow action dialog:
            # the dialog resize too and it gives a good feeling :P

        elif e.type() == QtCore.QEvent.MouseButtonRelease and e.buttons() & self.buttons:
            self.down_pos = None
            self.down_size = None
            if not self._has_resized:
                self.target.setMinimumSize(self._mns)
                self.target.setMaximumSize(self._mxs)
                self.target.setSizePolicy(self._sp)
                self._do_on_resized()


        return False




from warnings import warn

from .interface import Editor_Interface

from qtpy import QtWidgets, QtGui, QtCore


class ChoiceValueEditor(QtWidgets.QComboBox, Editor_Interface):
    '''
    This editor lets you choose one items in a choice list.

    Editor Type Names:
        choice
        choices (deprecated)

    Options:
        sorted:         bool (False) - sorts the choices
        choice_icons:   dict mapping each choice to an icon_ref
                        example: choice_icons={'Work in Progress':('icons.status', 'WIP')}
    '''
    @classmethod
    def can_edit(cls, editor_type_name):
        '''
        Must be implemented to return True if the given editor_type_name
        matches this editor.
        '''
        if editor_type_name in ('choices',):
            warn('Editor Type Name {!r} is deprecated.'.format(editor_type_name))
        return editor_type_name in ('choice', 'choices')  # Flow1 used choices... :/

    def __init__(self, parent, options):
        QtWidgets.QComboBox.__init__(self, parent)
        Editor_Interface.__init__(self, parent)
        self._sorted = False
        self._choice_icons = {}
        self.apply_options(options)

        self.activated.connect(self._on_select)

    def needs_choices(self):
        '''
        Must be overriden by editor presenting a choice of possible values.
        '''
        return True

    def set_editable(self, b):
        '''
        Must be implemented to prevent editing if b is False.
        Visual cue show also be given to the user.
        '''
        self.setEnabled(b)

    def apply_options(self, options):
        '''
        Must be implemented to configure the editor as
        described by the options dict.
        '''
        self._sorted = options.get('sorted', False)
        self._choice_icons = options.get('choice_icons', {})

    def update(self):
        '''
        Must be implemnented to show the value returned by self.fetch_value()
        Your code should call self._on_updated() at the end.
        '''
        value, choices = self.fetch_value()  # this is because needs_choices() returns True
        if self._sorted:
            choices.sort()

        self.clear()
        for i, choice in enumerate(choices):
            if choice is None:
                label = ''
            else:
                label = str(choice)

            try:
                icon_ref = self._choice_icons[choice]
            except KeyError:
                icon = self.get_icon_for(label)
            else:
                icon = self.get_icon_by_ref(icon_ref)

            if icon is not None:
                self.addItem(icon, label)
            else:
                self.addItem(label)
            if choice == value:
                self.setCurrentIndex(i)

        self._on_updated()

    def _on_select(self):
        self.apply()

    def get_edited_value(self):
        '''
        Must be implemented to return the value currently displayed.
        '''
        return self.currentText()

    def _show_edited(self):
        '''
        Must be implemented to show that the displayed value
        needs to be applied.
        '''
        # This will not happend, the changes are directly applied.
        self.setProperty('edited', True)
        self.setProperty('applying', False)
        self.style().polish(self)
        self.setText('Edited...')

    def _show_applied(self):
        '''
        Must be implemented to show that the displayed value
        as been saved.
        In a clean scenario, applying edits will trigger an update()
        and this state should disapear.
        If you are using the Editor without this kind of round trip,
        you can call update here.
        '''
        self.setProperty('applying', True)

    def _show_clean(self):
        '''
        Must be implemented to show that the displayed value is 
        up to date.
        '''
        self.setProperty('edited', False)
        self.setProperty('applying', False)
        self.style().polish(self)

    def _show_error(self, error_message):
        '''
        Must be implemented to show that the given error occured.
        '''
        self.setProperty('error', True)
        self.style().polish(self)
        self.setToolTip('/!\\ ERROR: %s' % error_message)



import re
import logging

from qtpy import QtWidgets

import kabaret.app.resources as resources


class Style(object):

    def __init__(self, name):
        super(Style, self).__init__()
        StylesManager.get().register_style(name, self)
        self._properties = {}

    def apply(self, widget=None):
        raise NotImplementedError()

    def apply_css(self, widget, css):
        """
        Set the css of the widget and load resources if referenced
        You can call a resource from the css with bracket and registered path of the resource
        Example:
        image: url({icons.gui.check});
        """
        regex = r'\{([\w\.\-_]+)\}'
        pattern = re.compile(regex)
        for match in pattern.finditer(css):
            path = match.group(1).split('.')
            try:
                css = css.replace(match.group(0), resources.get('.'.join(path[:-1]), path[-1])).replace('\\','/')
            except (resources.ResourcesError, resources.NotFoundError):
                logging.getLogger('kabaret.style').warning('Resource %r not found' % match.group(1))
                css = css.replace(match.group(0), 'none')
        widget.setStyleSheet(css)

    def set_property(self, name, value):
        self._properties[name] = value

    def get_property(self, name, default=None):
        return self._properties.get(name, default)


class StylesManager(object):

    _THE = None

    @classmethod
    def get(cls):
        return cls._THE

    def __init__(self):
        if self.__class__._THE is not None:
            raise Exception('The StylesManager is a singleton, you cant instantiate another one.')
        self.__class__._THE = self
        super(StylesManager, self).__init__()
        self._styles = {}
        self._default_style_name = None

    def register_style(self, name, style, set_default=True):
        self._styles[name] = style
        if set_default or self._default_style_name is None:
            self._default_style_name = name

    def get_style(self, name):
        return self._styles[name]

    def get_default_style(self):
        return self.get_style(self._default_style_name)

    def apply_default_style(self, widget=None):
        if self._default_style_name is None:
            raise Exception('No style registered !')
        self.apply_style(self._default_style_name, widget)

    def apply_style(self, style_name, widget=None):
        style = self.get_style(style_name)
        logging.getLogger('kabaret.style').debug(str(style))
        style.apply(
            widget or QtWidgets.QApplication.instance()
        )

StylesManager() # instanciate the singleton.





'''

    kabaret.flow.relations

    Defines the relations classes used to bind objects together:
        Parent:     Returns the parent of the object.
                    If the nb_levels argument is more than 1, a grand parent
                    is returned: 
                        1 -> parent
                        2 -> parent's parent
                        3 -> parent's parent's parent
                        etc...

        Relative:   Returns another Object in the flow.
                    The related Object is NOT a child of this Object, it must 
                    exist as a Child of another Object.

        Child:      Declares a Child in the Object. The Object will be the parent
                    of the created child.

        Param:      Declares a Child Value in the Object.
                    The default_value and value_type arguments control the initial data
                    returned by the child Value and the type of the child Value.
                    (a subclass of kabaret.flow.values.Value())
        
        SessionParam: Declare a Child SessionValue in the Object.
                    This value will be reset to its default at each session and
                    is not shared with other sessions.

        IntParam
        FloatParam
        StringParam
        DictParam
        OrderedStringSetParam
        HashParam:  Declares as Child the corresponding Value subclass.

        Separator:
                    A pre-configured Param acting as a separator in GUI

        Computed:   Declares a Child ComputedValue in the Object. 
                    See kabaret.flow.values.ComputedValue().

        Connection: Declares a Child Ref in the Object.
                    The ref being a Value returning another Object, this relations
                    is used to define dependencies between Objects.
                    See kabaret.flow.values.Ref()

'''
import logging
from .exceptions import (
    WIPException, 
    MissingChildError, MissingRelationError, 
)

from .object import _Relation
from .injection import injection

from .values import (
    Value, SessionValue,
    IntValue, BoolValue, FloatValue, StringValue, DictValue,
    OrderedStringSetValue, HashValue,
    ComputedValue, Ref
)


class Parent(_Relation):
    '''
    The Parent relation give access to the related object's parent or grand-parents.
    '''
    _RELATION_TYPE_NAME = 'Parent'

    def __init__(self, nb_levels=1):
        super(Parent, self).__init__(None)
        self.nb_levels = nb_levels

    def get_ui(self, of=None):
        '''
        Overridden to return the ui of the parented Object.
        See the `_Relation.get_ui()` class in `flow.object`.
        '''
        ui = super(Parent, self).get_ui(None)
        if of is not None:
            related = self.__get__(of)
            related._fill_ui(ui)
        return ui

    def __get__(self, o, t=None):
        if o is None:
            return self
        parent = o
        for i in range(self.nb_levels):
            if parent is None:
                raise Exception(
                    'Could not find a parent at level %d for %r' % (i, o._mng.oid()))
            parent = parent._mng.parent
        return parent

class Relative(_Relation):
    '''
    The Relative relation gives access to an Object which is not a child of the
    owner of the relation.
    GUI will display the Related object as a child though.
    
    The related object is selected by giving an oid relative to the relation owner.
    (an absolute oid can also be used, but pointing to the project iteself will fail)

    You can also subclass `Relative.get_relative_oid()` to programatically select
    the related Object, but beware: GUI wont update when the related oid changes unless
    the user refresh the page(s) displaying the Relative relation.
    
    You can use `Relative(...).ui(...)` to change the icon, editor, group, tooltip etc...
    of the related Object, but hidden=True will not work.
    You must use None as relative_oid to hide the relation in the GUI.
    '''

    _RELATION_TYPE_NAME = 'Relative'

    def __init__(self, relative_oid):
        '''
        The `relative_oid` is an oid relative to the relation owner.
        (An absolute oid is valide too)
        If None is used, the relation will be hidden.
        '''
        super(Relative, self).__init__(None)
        self._relative_oid = relative_oid

    def get_relative_oid(self, owner):
        '''
        Returns the relative oid given in the constructor.

        Subclasses can override this to programaticaly set
        the related Object (beware that GUI will not update
        when the return value changes though)

        If the return value is None, the relation will be
        hidden.

        The owner parameter is the Object owning this relation.
        You may use it to access other Objects in the flow when
        building the return value.
        '''
        return self._relative_oid

    def get_related(self, of):
        relative_oid = self.get_relative_oid(of)
        if relative_oid is None:
            return None

        try:
            related = of._mng.get_object(relative_oid)
        except (MissingChildError, MissingRelationError) as err:
            return None
        return related

class Child(_Relation):
    '''
    The Child relation sets an Object as the Child of the owner of the relation
    (wich in turn becomes the parent)
    '''
    _RELATION_TYPE_NAME = 'Child'

    def injectable(self, inherit_default=True):
        '''
        Declares an injection slot for this relation's related type.
        `inherit_default` defaults to True which will assert the injected
        type inherits the relation's related type.

        Injection resolution is done only once per type per project, 
        so you must use a different type (a simple subclass is enough)
        if you want the providers to be able to inject different types
        on different relations relating to the same type.
        '''
        injection.injectable(self.related_type, inherit_default)
        return self

    def _create_object(self, parent):
        related_type = injection.resolve(
                self.related_type, parent
            )
        try:
            related = related_type(parent, self.name)
        except:
            logging.getLogger('kabaret.flow').error(
                'Creating related object: %s(%r, %r)' % (
                    self.related_type.__name__, parent.oid(), self.name
                )
            )
            raise
        return related


class Param(Child):
    '''
    A Child relating to a :any:`Value` or one of its subclasses.
    '''

    _RELATION_TYPE_NAME = 'Param'

    _DEFAULT_VALUE_TYPE = Value

    def __init__(self, default_value=None, value_type=None):
        '''
        Beware of the default_value: if it is passed by reference, all
        instance of the class owning the created value will share the value!
        For example:
            class MyObject(Object):
                my_dict = Param({})

            -> all instances of MyObject will have their my_dict Value sharing
            the same dict (initialized when python executed the class definition)

        The workaround is to use a callable that creates a new instance of the value:
            class MyObject(Object):
                my_dict = Param(dict)

        '''
        if value_type is None:
            value_type = self.__class__._DEFAULT_VALUE_TYPE
        elif not issubclass(value_type, self.__class__._DEFAULT_VALUE_TYPE):
            raise TypeError(
                '%r is not a subclass or %r, needed for relation %r' % (
                    value_type, self.__class__._DEFAULT_VALUE_TYPE,
                    self
                )
            )

        super(Param, self).__init__(value_type)
        if callable(default_value):
            self.get_default_value = default_value
        else:
            self.get_default_value = lambda: default_value

        # default for Param is editable:
        self._ui['editable'] = True
        try:
            default_editor = value_type.DEFAULT_EDITOR
        except Exception:
            pass
        else:
            self._ui['editor_type'] = default_editor
        self._watched = False

    def watched(self, b=True):
        '''
        Configures the related value to be watched or not (default is False).
        Watched value call their parent's child_value_changed() when changed.
        '''
        self._watched = b
        return self

    def _create_object(self, parent):
        value = super(Param, self)._create_object(parent)
        value.set_default_value(self.get_default_value())
        value.set_watched(self._watched)
        return value

#---OLD
    def __set__(self, o, v):
        raise WIPException('Never used this, it is confusing.')
        # we must use get() on related values so not using set() feels strange
        # (you may end up thinking it's an attribute and not a related value...)
        value = self.get_related(o)
        value.set(v)


class SessionParam(Param):
    '''
    A Param relating to a :any:`SessionValue`.
    '''

    _DEFAULT_VALUE_TYPE = SessionValue


class IntParam(Param):
    '''
    A Param relating to an :any:`IntValue`.
    '''

    _DEFAULT_VALUE_TYPE = IntValue


class BoolParam(Param):
    '''
    A Param relating to a :any:`BoolValue`.
    '''

    _DEFAULT_VALUE_TYPE = BoolValue


class FloatParam(Param):
    '''
    A Param relating to a :any:`FloatValue`.
    '''

    _DEFAULT_VALUE_TYPE = FloatValue


class StringParam(Param):
    '''
    A Param relating to a :any:`StringValue`.
    '''

    _DEFAULT_VALUE_TYPE = StringValue


class DictParam(Param):
    '''
    A Param relating to a :any:`DictValue`.
    '''

    _DEFAULT_VALUE_TYPE = DictValue


class OrderedStringSetParam(Param):
    '''
    A Param relating to an :any:`OrderedStringSetValue`.
    '''

    _DEFAULT_VALUE_TYPE = OrderedStringSetValue

    def __init__(self, value_type=None):
        super(OrderedStringSetParam, self).__init__(None, value_type)


class HashParam(Param):
    '''
    A Param relating to a :any:`HashValue`.
    '''

    _DEFAULT_VALUE_TYPE = HashValue

    def __init__(self, value_type=None):
        super(HashParam, self).__init__({}, value_type)


def Separator():
    '''
    Returns a Param relation showing an horizontal line in GUI.
    '''
    return Param(None).ui(editor='label', text='<hr>').ui(label='')


def Label(text, label=''):
    '''
    Returns a Param relation showing a (potentially html) text in GUI.
    '''
    return Param(None).ui(editor='label', text=text).ui(label=label)


class Computed(Param):
    '''
    A Param relating to a :any:`ComputedValue`

    The value computation is delegated to the parent's compute_child_value() method.
    The 'cached' and 'store_value' constuctor arguments will configure the ComputedValue.
    (See kabaret.flow.values.ComputedValue)

    You can use a subclass of ComputedValue by specifying computed_value_type in the constructor
    '''
    _RELATION_TYPE_NAME = 'Computed'

    _DEFAULT_VALUE_TYPE = ComputedValue

    def __init__(self, cached=False, store_value=False, computed_value_type=None):
        super(Computed, self).__init__(None, value_type=computed_value_type)
        self._cached = cached
        self._store_value = store_value
        self.ui(editable=False)

    def _create_object(self, parent):
        value = super(Computed, self)._create_object(parent)
        value.set_cached(self._cached)
        value.set_store_value(self._store_value)
        return value


class Connection(Child):
    '''
    A Child relating to a :any:`Ref` subclass.
    '''
    _RELATION_TYPE_NAME = 'Connection'

    def __init__(self, related_type=None, ref_type=None):
        '''
        One of related_type and ref_type must be given:
            - if ref_type is not None, it will be used to reference the connected object
            - if ref_type is None, related_type.ref_type() will be used

        Specifying the ref_type instead of the related_type can be usefull in 
        obscur situation but should be avoided unless you have a good reason.
        (I would explain but I cant remind why I used this in the past :p)
        '''
        if ref_type is None:
            if related_type is not None:
                ref_type = related_type.ref_type()
            # else:
            # This mean the type will be configured later (I hope!)
            # It it used for example to have a ref to the same class.
        super(Connection, self).__init__(ref_type)

        # default for Connection is editable:
        self._ui['editable'] = True
        self._watched = False
        self._allow_cross_project = False

    def _create_object(self, parent):
        value = super(Connection, self)._create_object(parent)
        value.set_watched(self._watched)
        value.set_allow_cross_project(self._allow_cross_project)
        return value

    def watched(self, b=True):
        '''
        Configures the related value to be watched or not (default is False).
        Watched value call their parent's child_value_changed() when changed.
        '''
        self._watched = b
        return self
    
    def allow_cross_project(self, b=True):
        '''
        Configures the related Ref to accept a source belonging to another project.
        This is forbidden by default and you should be aware of the consequences
        before usign this (Hint: it related to your ability to backup/archive projects
        separately)
        '''
        self._allow_cross_project = b
        return self

#---OLD
    def __set__(self, o, v):
        raise WIPException('Never used this, it is confusing.')
        # we must use get() on related values so not using set() feels strange
        # (you may end up thinking it's an attribute and not a related value...)
        value = self.get_related(o)
        value.set(v)

#---OLD


class PipelineConnection(Connection):
    '''
    A Connection with ui(editable=False).
    '''

    def __init__(self, related_type=None, ref_type=None):
        raise WIPException('THIS IS OBSOLETE')
        super(PipelineConnection, self).__init__(related_type, ref_type)
        # a PipelineConnection is just a not GUI editable connection
        self._ui['editable'] = False
